﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using EFramework.Core.Events;
using Moq;
using NUnit.Framework;

namespace EFramework.Core.Tests.Events
{
    [TestFixture]
    public class EventsManagerFixture
    {
        #region Constructor Tests

        [Test]
        public void CanCreateNewInstance()
        {
            var sender = new object();
            var instance = new EventsManager(sender);

            Assert.NotNull(instance);
        }

        [Test]
        public void IsNotGuardingNullSenderWhenCreatingNewInstance()
        {
            var sender = default(object);
            var instance = new EventsManager(sender);
        }

        #endregion

        #region Add Tests

        [Test]
        public void CanAddDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = "Abc";
            var eventHandler = new Action(() => { });

            instance.Add(eventName, eventHandler);
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardNullEventNameWhenAddingDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = default(string);
            var eventHandler = new Action(() => { });

            instance.Add(eventName, eventHandler);
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardEmptyEventNameWhenAddingDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = String.Empty;
            var eventHandler = new Action(() => { });

            instance.Add(eventName, eventHandler);
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardNullDelegateWhenAddingDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = "Abc";
            var eventHandler = default(Delegate);

            instance.Add(eventName, eventHandler);
        }

        #endregion

        #region Remove Tests

        [Test]
        public void CanRemoveDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = "Abc";
            var eventHandler = new Action(() => { });

            instance.Remove(eventName, eventHandler);
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardNullEventNameWhenRemovingDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = default(string);
            var eventHandler = new Action(() => { });

            instance.Remove(eventName, eventHandler);
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardEmptyEventNameWhenRemovingDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = String.Empty;
            var eventHandler = new Action(() => { });

            instance.Remove(eventName, eventHandler);
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardNullDelegateWhenRemovingDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = "Abc";
            var eventHandler = default(Delegate);

            instance.Remove(eventName, eventHandler);
        }

        #endregion

        #region Invoke Tests

        [Test]
        public void CanInvokeAddedHandler()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = "Abc";
            var eventCounter = 0;
            var eventArgs = new EventArgs();
            var eventHandler = new Action<object, EventArgs>((_, args) => { if (Object.ReferenceEquals(args, eventArgs)) eventCounter++; });

            instance.Add(eventName, eventHandler);
            instance.Invoke(eventName, eventArgs);

            Assert.AreEqual(1, eventCounter, "Event should be once raised.");
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardNullEventNameWhenInvokingDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = default(string);
            var eventArgs = new EventArgs();

            instance.Invoke(eventName, eventArgs);
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardEmptyEventNameWhenInvokingDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = String.Empty;
            var eventArgs = new EventArgs();

            instance.Invoke(eventName, eventArgs);
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardNullEventArgsWhenInvokingDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = "Abc";
            var eventArgs = default(EventArgs);

            instance.Invoke(eventName, eventArgs);

        }

        #endregion

        #region BeginInvoke Tests

        [Test]
        public void CanBeginInvokeAddedHandler()
        {
            var sender = new object();
            var trigger = new Mock<EventsTrigger>();
            var instance = new EventsManager(sender, trigger.Object);

            var counter = 0;
            var eventName = "Abc";
            var eventArgs = new EventArgs();
            var eventHandler = new Action<object, EventArgs>((_, args) => { });

            trigger.Setup(x => x.BeginInvoke<EventArgs>(It.Is<IEnumerable<Delegate>>(e => e.Count() == 1), sender, eventArgs)).Callback(() => counter++).Verifiable();

            instance.Add(eventName, eventHandler);
            instance.BeginInvoke(eventName, eventArgs);

            Assert.AreEqual(1, counter, "The trigger should be called once.");
            trigger.VerifyAll();
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardNullEventNameWhenBeginInvokingDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = default(string);
            var eventArgs = new EventArgs();

            instance.BeginInvoke(eventName, eventArgs);
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardEmptyEventNameWhenBeginInvokingDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = String.Empty;
            var eventArgs = new EventArgs();

            instance.BeginInvoke(eventName, eventArgs);
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardNullEventArgsWhenBEginInvokingDelegate()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = "Abc";
            var eventArgs = default(EventArgs);

            instance.BeginInvoke(eventName, eventArgs);

        }

        #endregion

        #region Various Tests

        [Test]
        public void CanAddAndRemoveHandler()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = "Abc";
            var eventCounter = 0;
            var eventHandler = new Action<object, EventArgs>((_, args) => { eventCounter++; });

            instance.Add(eventName, eventHandler);
            instance.Remove(eventName, eventHandler);
            instance.Invoke(eventName, EventArgs.Empty);

            Assert.AreEqual(0, eventCounter, "The event should be triggered because it was removed.");
        }

        [Test]
        public void CanSubscribeTwiceTheSameEvent()
        {
            var sender = new object();
            var instance = new EventsManager(sender);
            var eventName = "Abc";

            var eventCounter1 = 0;
            var eventHandler1 = new Action<object, EventArgs>((_, args) => { eventCounter1++; });
            var eventCounter2 = 0;
            var eventHandler2 = new Action<object, EventArgs>((_, args) => { eventCounter2++; });

            instance.Add(eventName, eventHandler1);
            instance.Add(eventName, eventHandler2);
            instance.Invoke(eventName, EventArgs.Empty);

            Assert.AreEqual(1, eventCounter1, "The event should be triggered once.");
            Assert.AreEqual(1, eventCounter2, "The event should be triggered once.");
        }

        #endregion
    }
}
