﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using EFramework.Model.Tracking;
using NUnit.Framework;

namespace EFramework.Model.Tests
{
    [TestFixture]
    public class EntityFixture
    {
        #region Constructor Tests

        [Test]
        public void CanCreateNewInheritedInstance()
        {
            var id = Guid.NewGuid();
            var instance = new EntityA(id);

            Assert.NotNull(instance.Identifier, "Instance should properly set its identifier.");
            Assert.NotNull(instance.Tracker, "Instance should properly set its tracker.");
            Assert.AreEqual(id, instance.Identifier.Id, "Instance should have provided id.");
        }

        [Test]
        public void IsEntityAfterCreationMarkedAsAdded()
        {
            var instance = new EntityA(Guid.NewGuid());

            Assert.AreEqual(State.Added, instance.Tracker.State, "Instance should be marked as added");
        }

        #endregion

        #region Add and Remove Tests

        [Test]
        public void CanConnectTwoEntities()
        {
            var instanceA = new EntityA(Guid.NewGuid());
            var instanceB = new EntityB(Guid.NewGuid());

            instanceA.Add(instanceB.Identifier);

            CollectionAssert.IsNotEmpty(instanceA.GetAll(), "Instance should provide elements.");
            CollectionAssert.AllItemsAreNotNull(instanceA.GetAll(), "Instance should provide not null elements.");
            CollectionAssert.Contains(instanceA.GetAll(), instanceB.Identifier, "Instance should provide newly added element.");
        }

        [Test]
        public void CanDisconnectTwoEntities()
        {
            var instanceA = new EntityA(Guid.NewGuid());
            var instanceB = new EntityB(Guid.NewGuid());

            instanceA.Add(instanceB.Identifier);
            instanceA.Remove(instanceB.Identifier);

            CollectionAssert.IsEmpty(instanceA.GetAll(), "Instance should not provide any elements.");
        }

        [Test]
        public void CanConnectMoreEntitiesAtOnce()
        {
            var instanceA = new EntityA(Guid.NewGuid());
            var instanceB1 = new EntityB(Guid.NewGuid());
            var instanceB2 = new EntityB(Guid.NewGuid());

            instanceA.Add(instanceB1.Identifier, instanceB2.Identifier);

            CollectionAssert.IsNotEmpty(instanceA.GetAll(), "Instance should provide elements.");
            CollectionAssert.AllItemsAreNotNull(instanceA.GetAll(), "Instance should provide not null elements.");
            CollectionAssert.Contains(instanceA.GetAll(), instanceB1.Identifier, "Instance should provide newly added element.");
            CollectionAssert.Contains(instanceA.GetAll(), instanceB2.Identifier, "Instance should provide newly added element.");
        }

        [Test]
        public void CanConnectMoreDifferentEntitiesAtOnce()
        {
            var instanceA = new EntityA(Guid.NewGuid());
            var instanceB = new EntityB(Guid.NewGuid());
            var instanceC = new EntityC(Guid.NewGuid());

            instanceA.Add(instanceB.Identifier, instanceC.Identifier);

            CollectionAssert.IsNotEmpty(instanceA.GetAll(), "Instance should provide elements.");
            CollectionAssert.AllItemsAreNotNull(instanceA.GetAll(), "Instance should provide not null elements.");
            CollectionAssert.Contains(instanceA.GetAll(), instanceB.Identifier, "Instance should provide newly added element.");
            CollectionAssert.Contains(instanceA.GetAll(), instanceC.Identifier, "Instance should provide newly added element.");
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardIdentifiersWhenConnectingEntities()
        {
            var instance = new EntityA(Guid.NewGuid());

            instance.Add(null);
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardIdentifiersWhenDisconnectingEntities()
        {
            var instance = new EntityA(Guid.NewGuid());

            instance.Remove(null);
        }

        #endregion

        #region Searching Tests

        [Test]
        public void CanGetSingleReference()
        {
            var instanceA = new EntityA(Guid.NewGuid());
            var instanceB = new EntityB(Guid.NewGuid());

            instanceA.Add(instanceB.Identifier);
            var foundReference = instanceA.GetSingle<EntityB>();

            Assert.NotNull(foundReference, "The reference should be found.");
        }

        [Test]
        public void CanGetSingleOrDefaultReference()
        {
            var instanceA = new EntityA(Guid.NewGuid());
            var instanceB = new EntityB(Guid.NewGuid());

            instanceA.Add(instanceB.Identifier);
            var foundReference = instanceA.GetSingleOrDefault<EntityB>();

            Assert.NotNull(foundReference, "The reference should be found.");
        }
        
        [Test]
        public void CanHandleLookingForNotExistingSingleReferenceUsingSingleOrDefaultMethod()
        {
            var instanceA = new EntityA(Guid.NewGuid());
            var foundReference = instanceA.GetSingleOrDefault<EntityB>();

            Assert.Null(foundReference, "The reference should be not found.");
        }

        [Test, ExpectedException(typeof(KeyNotFoundException))]
        public void CanHandleLookingForNotExistingSingleReferenceUsingSingleMethod()
        {
            var instanceA = new EntityA(Guid.NewGuid());
            var foundReference = instanceA.GetSingle<EntityB>();
        }

        [Test]
        public void CanHandleLookignForSingleReferenceWhenMoreThanOnceExists()
        {
            var instanceA = new EntityA(Guid.NewGuid());
            var instanceB1 = new EntityB(Guid.NewGuid());
            var instanceB2 = new EntityB(Guid.NewGuid());

            instanceA.Add(instanceB1.Identifier, instanceB2.Identifier);
            var foundReference = instanceA.GetSingle<EntityB>();

            Assert.NotNull(foundReference, "The reference should be found.");
        }

        [Test]
        public void CanGetAllReferences()
        {
            var instanceA = new EntityA(Guid.NewGuid());
            var instanceB = new EntityB(Guid.NewGuid());
            var instanceC = new EntityC(Guid.NewGuid());

            instanceA.Add(instanceB.Identifier, instanceC.Identifier);

            CollectionAssert.Contains(instanceA.GetAll(), instanceB.Identifier, "Instance should provide newly added element.");
            CollectionAssert.Contains(instanceA.GetAll(), instanceC.Identifier, "Instance should provide newly added element.");
        }

        [Test]
        public void CanGetAllReferencesRequestType()
        {
            var instanceA = new EntityA(Guid.NewGuid());
            var instanceB = new EntityB(Guid.NewGuid());
            var instanceC = new EntityC(Guid.NewGuid());

            instanceA.Add(instanceB.Identifier, instanceC.Identifier);

            CollectionAssert.Contains(instanceA.GetAll<EntityB>(), instanceB.Identifier, "Instance should provide newly added element.");
            CollectionAssert.DoesNotContain(instanceA.GetAll<EntityB>(), instanceC.Identifier, "Instance should not provide not requested element.");
        }

        [Test]
        public void CanProvideEmptyReferenceListWhenThereIsNoElementsOfRequestedType()
        {
            var instanceA = new EntityA(Guid.NewGuid());
            var instanceB = new EntityB(Guid.NewGuid());

            instanceA.Add(instanceB.Identifier);

            CollectionAssert.IsEmpty(instanceA.GetAll<EntityC>(), "The reference list should be empty.");
        }
        
        #endregion

        #region Various Tests

        [Test]
        public void IsEntityAfterCreationAndAddingReferenceStillMarkedAsAdded()
        {
            var instanceA = new EntityA(Guid.NewGuid());
            var instanceB = new EntityB(Guid.NewGuid());

            instanceA.Add(instanceB.Identifier);

            Assert.AreEqual(State.Added, instanceA.Tracker.State, "Instance should be marked as added");
        }

        #endregion


        internal class EntityA : Entity
        {
            public EntityA(Guid id)
                : base(Identifier.From<EntityA>(id))
            {
            }
        }

        internal class EntityB : Entity
        {
            public EntityB(Guid id)
                : base(Identifier.From<EntityB>(id))
            {
            }
        }

        internal class EntityC : Entity
        {
            public EntityC(Guid id)
                : base(Identifier.From<EntityC>(id))
            {
            }
        }
    }
}
