﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using EFramework.Model.Tracking;
using NUnit.Framework;

namespace EFramework.Model.Tests.Tracking
{
    [TestFixture]
    public class TrackerFixture
    {
        #region Constructor

        [Test]
        public void CanCreateNewInstance()
        {
            var owner = new object();
            var instance = new Tracker(owner, State.Unchanged);

            Assert.AreEqual(owner, instance.Owner, "The owner property should have the assigned value.");
            Assert.AreEqual(State.Unchanged, instance.State, "The state should have the assigned value.");
            Assert.AreEqual(State.Unchanged, instance.StateInternal, "The state should have the assigned value.");
        }

        [Test]
        public void CanCreateNewInstanceWithChangedState()
        {
            var owner = new object();
            var instance = new Tracker(owner, State.Changed);

            Assert.AreEqual(owner, instance.Owner, "The owner property should have the assigned value.");
            Assert.AreEqual(State.Changed, instance.State, "The state should have the assigned value.");
            Assert.AreEqual(State.Changed, instance.StateInternal, "The state should have the assigned value.");
        }

        #endregion

        #region Assigning

        [Test]
        public void CanAssignAnotherTracker()
        {
            var instance1 = new Tracker(new object(), State.Unchanged);
            var instance2 = new Tracker(new object(), State.Unchanged);

            instance1.Add(instance2);

            Assert.AreEqual(State.Unchanged, instance1.State, "The state should remain unchanged.");
            Assert.AreEqual(State.Unchanged, instance1.StateInternal, "The state should remain unchanged.");
        }

        [Test]
        public void CanAssignAnotherChangedTracker()
        {
            var instance1 = new Tracker(new object(), State.Unchanged);
            var instance2 = new Tracker(new object(), State.Changed);

            instance1.Add(instance2);

            Assert.AreEqual(State.Changed, instance1.State, "The state should be changed.");
            Assert.AreEqual(State.Unchanged, instance1.StateInternal, "The state should remain unchanged.");
        }

        [Test]
        public void CanRemoveAnotherTracker()
        {
            var instance1 = new Tracker(new object(), State.Unchanged);
            var instance2 = new Tracker(new object(), State.Changed);

            instance1.Add(instance2);
            instance1.Remove(instance2);

            Assert.AreEqual(State.Unchanged, instance1.State, "The state should remain unchanged.");
            Assert.AreEqual(State.Unchanged, instance1.StateInternal, "The state should remain unchanged.");
        }

        [Test]
        public void CanClearTrackers()
        {
            var instance1 = new Tracker(new object(), State.Unchanged);
            var instance2 = new Tracker(new object(), State.Changed);

            instance1.Add(instance2);
            instance1.Clear();

            Assert.AreEqual(State.Unchanged, instance1.State, "The state should remain unchanged.");
            Assert.AreEqual(State.Unchanged, instance1.StateInternal, "The state should remain unchanged.");
        }

        [Test]
        public void CanAcceptChangedState()
        {
            var instance = new Tracker(new object(), State.Changed);

            instance.Accept();

            Assert.AreEqual(State.Unchanged, instance.State, "The state should be unchanged.");
            Assert.AreEqual(State.Unchanged, instance.StateInternal, "The state should be unchanged.");
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardChildTrackerWhenAddingIt()
        {
            var instance = new Tracker(new object(), State.Unchanged);

            instance.Add(null);
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardChildTrackerWhenRemovingIt()
        {
            var instance = new Tracker(new object(), State.Unchanged);

            instance.Remove(null);
        }

        #endregion

        #region Events

        [Test]
        public void CanRaiseStateChangedEventWhenInternalStateIsChanging()
        {
            var instance = new Tracker(new object(), State.Unchanged);
            var counter = 0;

            instance.StateChanged += (sender, args) => { counter++; };
            instance.Change(State.Changed);

            Assert.AreEqual(1, counter, "The event should be triggered.");
            Assert.AreEqual(State.Changed, instance.State, "The state should have the assigned value.");
            Assert.AreEqual(State.Changed, instance.StateInternal, "The state should have the assigned value.");
        }

        [Test]
        public void CanRaiseStateChangedEventWhenPropertyIsChanged()
        {
            var instance = new Tracker(new object(), State.Unchanged);
            var counter = 0;
            var propertyName = "Abc";

            instance.StateChanged += (sender, args) => { counter++; };
            instance.Change(propertyName);

            Assert.AreEqual(1, counter, "The event should be triggered.");
            Assert.AreEqual(State.Changed, instance.State, "The state should have the assigned value.");
            Assert.AreEqual(State.Changed, instance.StateInternal, "The state should have the assigned value.");
        }

        [Test]
        public void CanProperlySetTrackerEventArgsWhenPropertIsChanged()
        {
            var owner = new object();
            var instance = new Tracker(owner, State.Unchanged);
            var propertyName = "Abc";

            var raiserOwner = default(object);
            var raisedPropertyName = default(string);

            instance.Changed += (sender, args) => { raiserOwner = args.Owner; raisedPropertyName = args.PropertyName; };
            instance.Change(propertyName);

            Assert.AreEqual(owner, raiserOwner, "The owner should be properly set.");
            Assert.AreEqual(propertyName, raisedPropertyName, "The property name should be properly set.");
        }

        [Test]
        public void CanRaiseStateChangedEventWhenChildPropertyIsChanged()
        {
            var instance1 = new Tracker(new object(), State.Unchanged);
            var instance2 = new Tracker(new object(), State.Unchanged);

            instance1.Add(instance2);

            var counter = 0;
            var propertyName = "Abc";

            instance1.StateChanged += (sender, args) => { counter++; };
            instance2.Change(propertyName);

            Assert.AreEqual(1, counter, "The event should be triggered.");
            Assert.AreEqual(State.Changed, instance1.State, "The state should have the assigned value.");
            Assert.AreEqual(State.Unchanged, instance1.StateInternal, "The state should have the assigned value.");
        }

        [Test]
        public void CanRaiseStateChangedEventWhenStateWasAccepted()
        {
            var instance = new Tracker(new object(), State.Changed);
            var counter = 0;

            instance.StateChanged += (sender, args) => { counter++; };
            instance.Accept();

            Assert.AreEqual(1, counter, "The event should be triggered.");
            Assert.AreEqual(State.Unchanged, instance.State, "The state should have the assigned value.");
            Assert.AreEqual(State.Unchanged, instance.StateInternal, "The state should have the assigned value.");
        }
        #endregion
    }
}
