﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using EFramework.Core;
using EFramework.Core.Contracts;
using EFramework.Model.Tracking;

namespace EFramework.Model
{
    public abstract class Entity : SynchronizableObject, ICloneable
    {
        private readonly Identifier identifier;

        private readonly Dictionary<Type, HashSet<Identifier>> entries;
        private readonly HashSet<Identifier> references;

        private readonly Tracker tracker;
        private readonly Dictionary<Identifier, Tracker> trackers;


        protected Entity(Identifier identifier)
        {
            Ensure.IsNotNull(() => identifier);

            this.identifier = identifier;
            this.tracker = new Tracker(this, State.Added);
            this.entries = new Dictionary<Type, HashSet<Identifier>>();
            this.trackers = new Dictionary<Identifier, Tracker>();
            this.references = new HashSet<Identifier>();
        }


        public Identifier Identifier
        {
            get { return this.identifier; }
        }

        public Tracker Tracker
        {
            get { return this.tracker; }
        }

        internal Tracker GetTracker(Identifier identifier)
        {
            Ensure.IsNotNull(() => identifier);

            lock (this.Synchronizer)
            {
                return this.trackers[identifier];
            }
        }


        public void Add(params Identifier[] identifiers)
        {
            Ensure.IsNotNull(() => identifiers);

            this.Add(identifiers.AsEnumerable(), State.Added);
        }

        internal void Add(IEnumerable<Identifier> identifiers)
        {
            Ensure.IsNotNull(() => identifiers);

            this.Add(identifiers, State.Added);
        }

        internal void Add(IEnumerable<Identifier> identifiers, State state)
        {
            Ensure.IsNotNull(() => identifiers);

            lock (this.Synchronizer)
            {
                foreach (var identifier in identifiers)
                {
                    if (this.entries.ContainsKey(identifier.Owner) == false)
                    {
                        this.entries[identifier.Owner] = new HashSet<Identifier> { identifier };
                    }
                    else
                    {
                        this.entries[identifier.Owner].Add(identifier);
                    }

                    this.references.Add(identifier);
                    this.trackers.Add(identifier, new Tracker(null, state));
                    this.tracker.Add(this.trackers[identifier]);
                }
            }
        }

        public void Remove(params Identifier[] identifiers)
        {
            Ensure.IsNotNull(() => identifiers);

            lock (this.Synchronizer)
            {
                foreach (var identifier in identifiers)
                {
                    if (this.entries.ContainsKey(identifier.Owner) == true)
                    {
                        this.entries[identifier.Owner].Remove(identifier);
                    }

                    this.tracker.Remove(this.trackers[identifier]);
                    this.trackers.Remove(identifier);
                    this.references.Remove(identifier);
                }
            }
        }


        public Identifier GetSingle<TEntity>()
            where TEntity : Entity
        {
            lock (this.Synchronizer)
            {
                return this.entries[typeof(TEntity)].First();
            }
        }

        public Identifier GetSingleOrDefault<TEntity>()
            where TEntity : Entity
        {
            lock (this.Synchronizer)
            {
                if (this.entries.ContainsKey(typeof(TEntity)) == true)
                {
                    return this.entries[typeof(TEntity)].FirstOrDefault();
                }

                return default(Identifier);
            }
        }

        public ICollection<Identifier> GetAll()
        {
            lock (this.Synchronizer)
            {
                return this.references.ToArray();
            }
        }

        public ICollection<Identifier> GetAll<TEntity>()
            where TEntity : Entity
        {
            lock (this.Synchronizer)
            {
                if (this.entries.ContainsKey(typeof(TEntity)))
                {
                    return this.entries[typeof(TEntity)];
                }

                return new Identifier[] { };
            }
        }


        protected void OnChanged(string propertyName)
        {
            Ensure.IsNotNullNorEmpty(() => propertyName);

            lock (this.Synchronizer)
            {
                this.tracker.Change(propertyName);
            }
        }


        object ICloneable.Clone()
        {
            return this.Clone();
        }

        internal Entity Clone()
        {
            lock (this.Synchronizer)
            {
                var type = this.GetType();
                var instance = (Entity)Activator.CreateInstance(type, this.identifier.Id);
                var properties = from property in type.GetProperties()
                                 where property.DeclaringType == type
                                 select property;

                instance.Add(this.references);
                foreach (var property in properties)
                {
                    property.SetValue(instance, property.GetValue(this, null), null);
                }

                return instance;
            }
        }
    }
}
