﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using EFramework.Model.Tracking;
using EFramework.Model.Transfer;

namespace EFramework.Model.Serialization
{
    public class Serializer
    {
        public ContainerType Serialize(Container container, SerializerSettings settings)
        {
            var result = new ContainerType
            {
                Id = container.Identifier,
                Aliases = new List<AliasType>(),
                Entities = new List<EntityType>(),
            };

            var entityQuery = from entity in container.GetAll()
                              group entity by entity.Identifier.Owner into byTypes
                              select new
                              {
                                  Type = byTypes.Key,
                                  Entities = byTypes.ToArray(),
                              };

            foreach (var entry in entityQuery)
            {
                result.Aliases.Add(new AliasType { Code = entry.Type.Name, FullName = entry.Type.AssemblyQualifiedName });

                var propertyQuery = from property in entry.Type.GetProperties()
                                    where property.CanRead && property.CanWrite
                                    select property;

                var properties = propertyQuery.ToArray();

                foreach (var entity in entry.Entities)
                {
                    var entityType = new EntityType
                    {
                        Id = entity.Identifier.Id,
                        Type = entry.Type.Name,
                        Properties = new List<PropertyType>(),
                        References = new List<ReferenceType>(),
                    };

                    if (settings.PreserveTrackingStates)
                    {
                        entityType.TrackingState = (int)entity.Tracker.StateInternal;
                    }

                    var referenceQuery = from reference in entity.GetAll()
                                         group reference by reference.Owner into byTypes
                                         select new
                                         {
                                             Type = byTypes.Key,
                                             Entities = byTypes.ToArray(),
                                         };

                    foreach (var property in properties)
                    {
                        entityType.Properties.Add(new PropertyType
                        {
                            Name = property.Name,
                            Value = (string)Convert.ChangeType(property.GetValue(entity, null), typeof(string)),
                        });
                    }

                    foreach (var referenceEntry in referenceQuery)
                    {
                        foreach (var reference in referenceEntry.Entities)
                        {
                            var referenceType = new ReferenceType
                            {
                                Alias = referenceEntry.Type.Name,
                                Target = reference.Id,
                            };

                            if (settings.PreserveTrackingStates)
                            {
                                referenceType.TrackingState = (int)entity.GetTracker(reference).StateInternal;
                            }

                            entityType.References.Add(referenceType);
                        }
                    }

                    result.Entities.Add(entityType);
                }
            }

            return result;
        }

        public Container Deserialize(ContainerType containerType, SerializerSettings settings)
        {
            var container = new Container(containerType.Id);
            var entityQuery = from entity in containerType.Entities
                              group entity by entity.Type into byTypes
                              join alias in containerType.Aliases on byTypes.Key equals alias.Code
                              select new
                              {
                                  Type = Type.GetType(alias.FullName),
                                  Entities = byTypes.ToArray(),
                              };

            foreach (var entry in entityQuery)
            {
                var propertyQuery = from property in entry.Type.GetProperties()
                                    where property.CanRead && property.CanWrite
                                    select property;

                var properties = propertyQuery.ToArray();

                foreach (var entityType in entry.Entities)
                {
                    var entity = (Entity)Activator.CreateInstance(entry.Type, entityType.Id);
                    var propertyValueQuery = from property in properties
                                             join propertyType in entityType.Properties on property.Name equals propertyType.Name
                                             select new
                                             {
                                                 Property = property,
                                                 Value = Converter.ChangeType(property.PropertyType, propertyType.Value),
                                             };

                    foreach (var propertyEntry in propertyValueQuery)
                    {
                        propertyEntry.Property.SetValue(entity, propertyEntry.Value, null);
                    }

                    var referenceQuery = from reference in entityType.References
                                         group reference by reference.Alias into byTypes
                                         join alias in containerType.Aliases on byTypes.Key equals alias.Code
                                         select new
                                         {
                                             Type = Type.GetType(alias.FullName),
                                             References = byTypes.ToArray(),
                                         };

                    var identifierQuery = from referenceEntry in referenceQuery
                                          from referenceType in referenceEntry.References
                                          select new
                                          {
                                              Identifier = Identifier.From(referenceEntry.Type, referenceType.Target),
                                              Reference = referenceType,
                                          };

                    if (settings.PreserveTrackingStates)
                    {
                        entity.Tracker.Change((State)entityType.TrackingState);
                        identifierQuery = identifierQuery.ToArray();
                    }

                    entity.Add(identifierQuery.Select(x => x.Identifier), State.Unchanged);

                    if (settings.PreserveTrackingStates)
                    {
                        foreach (var reference in identifierQuery)
                        {
                            entity.GetTracker(reference.Identifier).Change((State)reference.Reference.TrackingState);
                        }
                    }

                    container.Add(entity);
                }
            }

            return container;
        }
    }
}
