﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using EFramework.Core.Contracts;
using EFramework.Core.Events;

namespace EFramework.Model.Tracking
{
    public class Tracker
    {
        private readonly object owner;
        private readonly HashSet<Tracker> trackers;
        private readonly EventsManager eventsManager;

        private State state;
        private bool changed;


        internal Tracker(object owner, State state)
        {
            this.owner = owner;
            this.trackers = new HashSet<Tracker>();
            this.eventsManager = new EventsManager(this);

            this.state = state;
            this.changed = false;
        }


        public object Owner
        {
            get { return this.owner; }
        }


        private const string ChangedEvent = "Changed";
        internal event EventHandler<TrackerEventArgs> Changed
        {
            add { this.eventsManager.Add(Tracker.ChangedEvent, value); }
            remove { this.eventsManager.Remove(Tracker.ChangedEvent, value); }
        }

        private void OnChanged(string propertyName)
        {
            Ensure.IsNotNullNorEmpty(() => propertyName);

            this.eventsManager.Invoke(Tracker.ChangedEvent, new TrackerEventArgs(this.owner, propertyName));
        }

        private void OnChanged(object sender, TrackerEventArgs e)
        {
            this.eventsManager.Invoke(Tracker.ChangedEvent, e);
        }


        private const string StateChangedEvent = "StateChanged";
        internal event EventHandler StateChanged
        {
            add { this.eventsManager.Add(Tracker.StateChangedEvent, value); }
            remove { this.eventsManager.Remove(Tracker.StateChangedEvent, value); }
        }

        private void OnStateChanged()
        {
            this.eventsManager.Invoke(Tracker.StateChangedEvent, EventArgs.Empty);
        }

        /// <summary>
        /// Gets the state of the tracker including the state of its dependants.
        /// </summary>
        public State State
        {
            get 
            {
                if (this.state != State.Unchanged)
                {
                    return this.state;
                }
                else if (this.changed == true)
                {
                    return State.Changed;
                }
                else
                {
                    return State.Unchanged;
                }
            }
        }

        /// <summary>
        /// Gets the internal state of the tracker.
        /// </summary>
        internal State StateInternal
        {
            get { return this.state; }
        }


        internal void Change(string propertyName)
        {
            Ensure.IsNotNullNorEmpty(() => propertyName);

            this.Change(State.Changed);
            this.OnChanged(propertyName);
        }

        internal void Change(State state)
        {
            if (this.state == State.Unchanged)
            {
                var old = this.State;
                this.state = state;

                if (old != this.State)
                    this.OnStateChanged();
            }
        }

        internal void Accept()
        {
            var old = this.State;

            if (this.state == State.Added || this.state == State.Changed)
            {
                this.state = State.Unchanged;
            }

            if (old != this.State)
                this.OnStateChanged();
        }


        private void OnStateChanged(object sender, EventArgs e)
        {
            var old = this.State;
            var tracker = sender as Tracker;
            var state = tracker.State;

            this.changed = state != State.Unchanged;

            if (this.changed == false)
                this.Recalculate();

            if (old != this.State)
                this.OnStateChanged();
        }

        private void Recalculate()
        {
            this.changed = false;

            foreach (var tracker in this.trackers)
            {
                if (tracker.State != State.Unchanged)
                {
                    this.changed = true;
                    break;
                }
            }
        }

        internal void Add(Tracker tracker)
        {
            Ensure.IsNotNull(() => tracker);

            var old = this.State;

            this.trackers.Add(tracker);
            tracker.StateChanged += this.OnStateChanged;
            tracker.Changed += this.OnChanged;

            if (this.changed == false)
                this.changed = tracker.State != State.Unchanged;

            if (old != this.State)
                this.OnStateChanged();
        }

        internal void Remove(Tracker tracker)
        {
            Ensure.IsNotNull(() => tracker);

            var old = this.State;

            this.trackers.Remove(tracker);
            tracker.StateChanged -= this.OnStateChanged;
            tracker.Changed -= this.OnChanged;

            if (this.changed == true)
                this.Recalculate();

            if (old != this.State)
                this.OnStateChanged();
        }

        internal void Clear()
        {
            var old = this.State;

            foreach (var tracker in this.trackers)
            {
                tracker.StateChanged -= this.OnStateChanged;
                tracker.Changed -= this.OnChanged;
            }

            this.trackers.Clear();
            this.Recalculate();

            if (old != this.State)
                this.OnStateChanged();
        }
    }
}
