﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Text;

namespace EFramework.Networking
{
    /// <summary>
    /// Represents the container with returned headers.
    /// </summary>
    public class NetworkCallback : INetworkModifier, IEnumerable
    {
        /// <summary>
        /// The null headers container.
        /// </summary>
        internal static readonly NetworkCallback Null = new NetworkCallback();

        /// <summary>
        /// Holds headers values.
        /// </summary>
        private readonly Dictionary<string, string> data;


        /// <summary>
        /// Initializes an empty instance of the <see cref="NetworkCallback"/> class.
        /// </summary>
        public NetworkCallback()
        {
            this.data = new Dictionary<string, string>();
        }


        /// <summary>
        /// Gets the header value.
        /// </summary>
        /// <param name="index">The header key.</param>
        /// <returns>The found header value or <see cref="String.Empty"/> value.</returns>
        public string this[string index]
        {
            get { return this.data.ContainsKey(index) ? data[index] : String.Empty; }
        }

        /// <summary>
        /// Adds a key/value pair to the callback dictionary.
        /// </summary>
        /// <param name="key">The key.</param>
        /// <param name="value">The value.</param>
        public void Add(string key, string value)
        {
            if (this != NetworkCallback.Null)
                this.data[key] = value;
        }

        /// <summary>
        /// Determines if the given key exists in the callback dictionary.
        /// </summary>
        /// <param name="key">The key to find.</param>
        /// <returns>True if the key exists in the dictionary; otherwise false.</returns>
        public bool Contains(string key)
        {
            return this.data.ContainsKey(key);
        }


        /// <summary>
        /// Modifies the <see cref="Uri"/> class.
        /// </summary>
        /// <param name="uri">The uri to modify.</param>
        /// <returns>The modified uri.</returns>
        Uri INetworkModifier.Modify(Uri uri)
        {
            return uri;
        }

        /// <summary>
        /// Modifies the <see cref="HttpWebRequest"/> class applying the POST data.
        /// </summary>
        /// <param name="request">The request to modify.</param>
        void INetworkModifier.Apply(HttpWebRequest request)
        {
        }

        /// <summary>
        /// Modifies the <see cref="HttpWebResponse"/> class.
        /// </summary>
        /// <param name="response">The response to modify.</param>
        void INetworkModifier.Apply(HttpWebResponse response)
        {
            if (this != NetworkCallback.Null)
            {
                foreach (var key in response.Headers.AllKeys)
                {
                    this.data[key] = response.Headers[key];
                }
            }
        }


        /// <summary>
        /// Do nothing. Forced implementation by dictionary class initializer.
        /// </summary>
        /// <returns>The null value.</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.data.GetEnumerator();
        }


        /// <summary>
        /// Returns the string that represents current object.
        /// </summary>
        /// <returns>The string representing current object.</returns>
        public override string ToString()
        {
            var builder = new StringBuilder();

            foreach (var header in this.data)
            {
                builder.Append(header.Key);
                builder.Append(": ");
                builder.AppendLine(header.Value);
            }

            return builder.ToString();
        }
    }
}
