﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;

namespace EFramework.Networking
{
    /// <summary>
    /// Represents the container with POST data to sent.
    /// </summary>
    public class NetworkGetData : INetworkModifier, IEnumerable
    {
        /// <summary>
        /// The empty GET data container.
        /// </summary>
        public static readonly NetworkGetData Empty = new NetworkGetData(false);

        /// <summary>
        /// The null GET data container.
        /// </summary>
        internal static readonly NetworkGetData Null = new NetworkGetData(true);

        /// <summary>
        /// Holds GEt data values.
        /// </summary>
        private readonly List<Entry> data;


        /// <summary>
        /// Initializes an empty instance of the <see cref="NetworkGetData"/> class.
        /// </summary>
        public NetworkGetData()
        {
            this.data = new List<Entry>();
        }

        /// <summary>
        /// Initializes an empty or null instance of the <see cref="NetworkGetData"/> class.
        /// </summary>
        /// <param name="isNull">True if the newly created instance should be NULL.</param>
        private NetworkGetData(bool isNull)
        {
            if (!isNull)
            {
                this.data = new List<Entry>();
            }
        }

        /// <summary>
        /// Initializes an instance of the <see cref="NetworkData"/> class.
        /// </summary>
        /// <param name="data">The dictionary to initialize with.</param>
        public NetworkGetData(IDictionary<string, string> data)
        {
            this.data = new List<Entry>();

            foreach (var entry in data)
            {
                this.data.Add(new Entry { Key = entry.Key, Value = entry.Value });
            }
        }


        /// <summary>
        /// Adds new value/key pair to the container.
        /// </summary>
        /// <param name="key">The key.</param>
        /// <param name="value">The value.</param>
        public void Add(string key, string value)
        {
            this.data.Add(new Entry { Key = key, Value = value });
        }


        /// <summary>
        /// Modifies the <see cref="Uri"/> class.
        /// </summary>
        /// <param name="uri">The uri to modify.</param>
        /// <returns>The modified uri.</returns>
        Uri INetworkModifier.Modify(Uri uri)
        {
            if (this.data != null)
            {
                var uriBuilder = new UriBuilder(uri);
                var queryBuilder = new StringBuilder();

                foreach (var entry in this.data)
                {
                    queryBuilder.Append(entry.Key);
                    queryBuilder.Append("=");
                    queryBuilder.Append(entry.Value);
                    queryBuilder.Append("&");
                }

                queryBuilder.Remove(queryBuilder.Length - 1, 1);
                uriBuilder.Query = queryBuilder.ToString();

                return uriBuilder.Uri;
            }
            else
            {
                return uri;
            }
        }

        /// <summary>
        /// Modifies the <see cref="HttpWebRequest"/> class applying the POST data.
        /// </summary>
        /// <param name="request">The request to modify.</param>
        void INetworkModifier.Apply(HttpWebRequest request)
        {
        }

        /// <summary>
        /// Modifies the <see cref="HttpWebResponse"/> class.
        /// </summary>
        /// <param name="response">The response to modify.</param>
        void INetworkModifier.Apply(HttpWebResponse response)
        {
        }


        /// <summary>
        /// Do nothing. Forced implementation by dictionary class initializer.
        /// </summary>
        /// <returns>The null value.</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.data.GetEnumerator();
        }


        /// <summary>
        /// Represents POST data key/value pair.
        /// </summary>
        internal struct Entry
        {
            public string Key;
            public string Value;
        }
    }
}
