﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Text;
using System.Reflection;

namespace EFramework.Networking
{
    /// <summary>
    /// Represents the container with headers to use.
    /// </summary>
    public class NetworkHeaders : INetworkModifier, IEnumerable
    {
        /// <summary>
        /// The null headers container.
        /// </summary>
        internal static readonly NetworkHeaders Null = new NetworkHeaders();

        /// <summary>
        /// Holds headers values.
        /// </summary>
        private readonly Dictionary<string, string> data;


        /// <summary>
        /// Initializes an empty instance of the <see cref="NetworkHeaders"/> class.
        /// </summary>
        public NetworkHeaders()
        {
            this.data = new Dictionary<string, string>();
        }

        /// <summary>
        /// Initializes an instance of the <see cref="NetworkHeaders"/> class.
        /// </summary>
        /// <param name="data">The dictionary to initialize with.</param>
        public NetworkHeaders(IDictionary<string, string> data)
        {
            this.data = new Dictionary<string, string>(data);
        }

        /// <summary>
        /// Adds new value/key pair to the container.
        /// </summary>
        /// <param name="key">The key.</param>
        /// <param name="value">The value.</param>
        public void Add(string key, string value)
        {
            this.data.Add(key, value);
        }


        /// <summary>
        /// Modifies the <see cref="Uri"/> class.
        /// </summary>
        /// <param name="uri">The uri to modify.</param>
        /// <returns>The modified uri.</returns>
        Uri INetworkModifier.Modify(Uri uri)
        {
            return uri;
        }

        /// <summary>
        /// Modifies the <see cref="HttpWebRequest"/> class applying the POST data.
        /// </summary>
        /// <param name="request">The request to modify.</param>
        void INetworkModifier.Apply(HttpWebRequest request)
        {
            var headersType = request.Headers.GetType();
            var headersMethod = headersType.GetMethod("SetAddVerified", BindingFlags.Instance | BindingFlags.NonPublic);

            foreach (var entry in this.data)
            {
                headersMethod.Invoke(request.Headers, new object[] { entry.Key, entry.Value });
            }
        }


        /// <summary>
        /// Modifies the <see cref="HttpWebResponse"/> class.
        /// </summary>
        /// <param name="response">The response to modify.</param>
        void INetworkModifier.Apply(HttpWebResponse response)
        {
        }


        /// <summary>
        /// Do nothing. Forced implementation by dictionary class initializer.
        /// </summary>
        /// <returns>The null value.</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return null;
        }
    }
}
