﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace EFramework.Networking
{
    public class NetworkLimiter
    {
        private readonly Dictionary<TimeSpan, int> definitions;
        private readonly List<RequestEntry> requests;
        private DateTime? lastCleanUp;
        private TimeSpan? cleanUpSpan;
        private TimeSpan? maxPeriod;

        private NetworkLimiter()
        {
            this.definitions = new Dictionary<TimeSpan, int>();
            this.requests = new List<RequestEntry>();
        }


        public static NetworkLimiter Create()
        {
            return new NetworkLimiter();
        }

        public NetworkLimiter Add(int seconds, int requests)
        {
            this.definitions[TimeSpan.FromSeconds(seconds)] = requests;
            this.cleanUpSpan = this.definitions.Keys.Min();
            this.maxPeriod = this.definitions.Keys.Max();

            return this;
        }


        internal void Register(string url)
        {
            var now = DateTime.Now;
            var host = String.Intern(new Uri(url).Host.ToLower());

            lock (this.requests)
            {
                this.CleanUp(now);
                this.requests.Add(new RequestEntry { Host = host, Timestamp = now });
            }
        }

        internal bool CanRequest(string url)
        {
            var now = DateTime.Now;
            var host = String.Intern(new Uri(url).Host.ToLower());

            lock (this.requests)
            {
                this.CleanUp(now);

                foreach (var period in this.definitions.Keys)
                    if (this.requests.Count(x => x.Host == host && x.Timestamp > now.Subtract(period)) > this.definitions[period])
                        return false;

                return true;
            }
        }

        private void CleanUp(DateTime now)
        {
            if (this.cleanUpSpan != null && this.lastCleanUp < now.Subtract(this.cleanUpSpan.Value))
            {
                this.lastCleanUp = now;
                this.requests.RemoveAll(x => x.Timestamp < now.Subtract(this.maxPeriod.Value));
            }
        }


        private class RequestEntry
        {
            public DateTime Timestamp;
            public string Host;
        }
    }
}
