﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;

namespace EFramework.Networking
{
    /// <summary>
    /// Represents the container with POST data to sent.
    /// </summary>
    public class NetworkPostData : INetworkModifier, IEnumerable
    {
        /// <summary>
        /// The empty POST data container.
        /// </summary>
        public static readonly NetworkPostData Empty = new NetworkPostData(false);

        /// <summary>
        /// The null POST data container.
        /// </summary>
        internal static readonly NetworkPostData Null = new NetworkPostData(true);

        /// <summary>
        /// Holds POST data values.
        /// </summary>
        private readonly List<Entry> data;


        /// <summary>
        /// Initializes an empty instance of the <see cref="NetworkPostData"/> class.
        /// </summary>
        public NetworkPostData()
        {
            this.data = new List<Entry>();
        }

        /// <summary>
        /// Initializes an empty or null instance of the <see cref="NetworkPostData"/> class.
        /// </summary>
        /// <param name="isNull">True if the newly created instance should be NULL.</param>
        private NetworkPostData(bool isNull)
        {
            if (!isNull)
            {
                this.data = new List<Entry>();
            }
        }

        /// <summary>
        /// Initializes an instance of the <see cref="NetworkPostData"/> class.
        /// </summary>
        /// <param name="data">The dictionary to initialize with.</param>
        public NetworkPostData(IDictionary<string, string> data)
        {
            this.data = new List<Entry>();

            foreach (var entry in data)
            {
                this.data.Add(new Entry { Key = entry.Key, Value = entry.Value });
            }
        }


        /// <summary>
        /// Adds new value/key pair to the container.
        /// </summary>
        /// <param name="key">The key.</param>
        /// <param name="value">The value.</param>
        public void Add(string key, string value)
        {
            this.data.Add(new Entry { Key = key, Value = value });
        }

        /// <summary>
        /// Gets the entry value.
        /// </summary>
        /// <param name="key">The entry key.</param>
        /// <returns>Found value or null.</returns>
        public string this[string key]
        {
            get { return this.data.FirstOrDefault(x => x.Key.Equals(key)).Value; }
        }


        /// <summary>
        /// Modifies the <see cref="Uri"/> class.
        /// </summary>
        /// <param name="uri">The uri to modify.</param>
        /// <returns>The modified uri.</returns>
        Uri INetworkModifier.Modify(Uri uri)
        {
            return uri;
        }

        /// <summary>
        /// Modifies the <see cref="HttpWebRequest"/> class applying the POST data.
        /// </summary>
        /// <param name="request">The request to modify.</param>
        void INetworkModifier.Apply(HttpWebRequest request)
        {
            if (this.data != null)
            {
                var builder = new StringBuilder();
                foreach (var entry in this.data)
                {
                    builder.Append(entry.Key);
                    builder.Append("=");
                    builder.Append(Uri.EscapeDataString(entry.Value));
                    builder.Append("&");
                }

                if (builder.Length > 0 && builder[builder.Length - 1] == '&')
                {
                    builder.Remove(builder.Length - 1, 1);
                }

                request.Method = WebRequestMethods.Http.Post;
                request.ContentType = @"application/x-www-form-urlencoded";

                using (var stream = request.GetRequestStream())
                using (var writer = new StreamWriter(stream))
                {
                    writer.Write(builder.ToString());
                    writer.Flush();
                    writer.Close();

                    stream.Flush();
                    stream.Close();
                }
            }
        }

        /// <summary>
        /// Modifies the <see cref="HttpWebResponse"/> class.
        /// </summary>
        /// <param name="response">The response to modify.</param>
        void INetworkModifier.Apply(HttpWebResponse response)
        {
        }


        /// <summary>
        /// Do nothing. Forced implementation by dictionary class initializer.
        /// </summary>
        /// <returns>The null value.</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.data.GetEnumerator();
        }

        
        /// <summary>
        /// Represents POST data key/value pair.
        /// </summary>
        internal struct Entry
        {
            public string Key;
            public string Value;
        }
    }
}
