﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading;
using EFramework.Networking.Parser;

namespace EFramework.Networking
{
    public class NetworkService
    {
        private readonly NetworkLimiter limiter;


        static NetworkService()
        {
            ServicePointManager.Expect100Continue = false;
        }


        public NetworkService(NetworkLimiter limiter = null)
        {
            this.limiter = limiter ?? NetworkLimiter.Create();
        }


        public NetworkGateway CreateGateway()
        {
            return new NetworkGatewayImpl(this);
        }


        internal Stream GetStream(NetworkRequest request)
        {
            this.EnsureLimited(request.Url);

            return this.GetStreamData(request.Url, request.GetModifiers());
        }

        private void EnsureLimited(string url)
        {
            while (true)
            {
                lock (this.limiter)
                {
                    if (this.limiter.CanRequest(url))
                    {
                        this.limiter.Register(url);
                        break;
                    }
                }

                Thread.Sleep(100);
            }
        }

        private Stream GetStreamData(string url, INetworkModifier[] modifiers)
        {
            var request = this.CreateRequest(url, modifiers);
            var response = this.GetResponse(request, modifiers);

            return response.GetResponseStream();
        }

        private HttpWebRequest CreateRequest(string url, INetworkModifier[] modifiers)
        {
            var uri = new Uri(url);

            foreach (var modifier in modifiers)
            {
                uri = modifier.Modify(uri);
            }

            var request = (HttpWebRequest)WebRequest.CreateDefault(uri);

            foreach (var modifier in modifiers)
            {
                modifier.Apply(request);
            }

            return request;
        }

        private HttpWebResponse GetResponse(HttpWebRequest request, params INetworkModifier[] modifiers)
        {
            HttpWebResponse response;

            try
            {
                response = (HttpWebResponse)request.GetResponse();
            }
            catch (WebException ex)
            {
                response = (HttpWebResponse)ex.Response;

                if (response == null)
                    throw;
            }

            foreach (var modifier in modifiers)
            {
                modifier.Apply(response);
            }

            return response;
        }
    }
}
