﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using EFramework.Networking.Parser;
using Majestic12;

namespace EFramework.Networking.Parser
{
    public static class HDocumentFactory
    {
        public static HDocument Create(string data)
        {
            using (var parser = new HTMLparser(data) { bDecodeEntities = true })
                return HDocumentFactory.Parse(parser);
        }

        private static HDocument Parse(HTMLparser parser)
        {
            // Without children
            var singles = new[] { "meta", "br", "input", "img", "area", "link", "embed" };
            // Ignoring children
            var childless = new[] { "style" };
            // With parent restrictions
            var under = new Dictionary<string, string[]> { { "tr", new[] { "table", "tbody", "thead", "tfoot", "form" } }, { "td", new[] { "tr", "table" } }, { "li", new[] { "ul", "ol" } } };
            // With close restrictions
            var close = new[] { "ul", "ol", "table", "p" };
            var stack = new Stack<HContainer>();
            var document = new HDocument();
            int indent = 0;

            stack.Push(document);

            while (true)
            {
                var tag = parser.ParseNext();
                if (tag == null)
                    break;

                switch (tag.oType)
                {
                    case HTMLchunkType.OpenTag: indent++;

                        if (stack.Count > 0 && stack.Peek() is HTag && singles.Contains((stack.Peek() as HTag).Tag))
                        {
                            stack.Pop();
                            indent--;
                        }

                        if (under.ContainsKey(tag.sTag))
                        {
                            while (stack.Count > 0 && stack.Peek() is HTag)
                            {
                                if (under[tag.sTag].Contains((stack.Peek() as HTag).Tag))
                                    break;

                                stack.Pop();
                                indent--;
                            }
                        }

                        if (stack.Count > 0 && stack.Peek() is HTag && childless.Contains((stack.Peek() as HTag).Tag))
                        {
                            break;
                        }

                        HDocumentFactory.Dump(new string(' ', indent));
                        HDocumentFactory.Dump(tag.sTag);

                        if (tag.bClosure)
                            indent--;

                        if (tag.sTag.Length > 0)
                        {
                            var created = new HTag(stack.Peek(), tag.sTag);

                            stack.Peek().Children.Add(created);
                            stack.Push(created);

                            if (tag.oParams != null)
                                foreach (string param in tag.oParams.Keys)
                                {
                                    created.Attributes.Add(new HAttribute(param, tag.GetParamValue(param)));
                                    HDocumentFactory.Dump(" " + param);
                                }

                            if (tag.bClosure)
                                stack.Pop();
                        }

                        HDocumentFactory.DumpLine(String.Empty);

                        break;
                    case HTMLchunkType.CloseTag:

                        if (!tag.bEndClosure)
                        {
                            if (stack.Count > 0 && stack.Peek() is HTag && (stack.Peek() as HTag).Tag != tag.sTag && singles.Contains((stack.Peek() as HTag).Tag))
                            {
                                stack.Pop();
                                indent--;
                            }

                            if (stack.Count > 0 && stack.Peek() is HTag && (stack.Peek() as HTag).Tag != tag.sTag && close.Contains(tag.sTag))
                            {
                                while (stack.Count > 0 && stack.Peek() is HTag)
                                {
                                    if ((stack.Peek() as HTag).Tag == tag.sTag)
                                        break;

                                    stack.Pop();
                                    indent--;
                                }
                            }

                            if (stack.Count > 0 && stack.Peek() is HTag && (stack.Peek() as HTag).Tag == tag.sTag)
                            {
                                indent--;
                                HDocumentFactory.Dump(new string(' ', indent));
                                HDocumentFactory.DumpLine("/" + tag.sTag);

                                stack.Pop();
                            }
                        }

                        break;
                    case HTMLchunkType.Script:

                        HDocumentFactory.Dump(new string(' ', indent + 1));
                        HDocumentFactory.DumpLine("script");
                        {
                            var created = (HScript)stack.Peek().Children.Add(new HScript(stack.Peek(), tag.oHTML));

                            if (tag.oParams != null)
                            {
                                foreach (string param in tag.oParams.Keys)
                                {
                                    created.Attributes.Add(new HAttribute(param, tag.GetParamValue(param)));
                                    HDocumentFactory.Dump(" " + param);
                                }
                            }
                        }


                        break;
                    case HTMLchunkType.Comment:

                        HDocumentFactory.Dump(new string(' ', indent + 1));
                        HDocumentFactory.DumpLine("comment");
                        stack.Peek().Children.Add(new HComment(stack.Peek(), tag.oHTML));

                        break;
                    case HTMLchunkType.Text:

                        if (tag.oHTML.Trim().Length > 0)
                            stack.Peek().Children.Add(new HText(stack.Peek(), tag.oHTML));

                        break;
                }
            }

            return document;
        }


        [Conditional("DUMP")]
        private static void Dump(string line)
        {
            Debug.Write(line);
        }

        [Conditional("DUMP")]
        private static void DumpLine(string line)
        {
            Debug.WriteLine(line);
        }
    }
}
