﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace EFramework.Networking.Parser
{
    public class HQuery
    {
        private readonly HContainer container;
        private readonly List<string> parts;
        private readonly List<HContainer> skips;

        private HQuery(HContainer document)
        {
            this.container = document;
            this.parts = new List<string>();
            this.skips = new List<HContainer>();
        }

        private void AddParts(string[] parts)
        {
            this.parts.AddRange(parts);
        }

        private IEnumerable<HTag> Execute()
        {
            var current = this.container is HTag ? new[] { this.container as HTag } : this.container.Children.OfType<HTag>();
            var children = false;
            var sibling = false;

            foreach (var part in this.parts)
            {
                if (part.Contains('.'))
                {
                    var fragments = part.Split(new[] { '.' }, 2);
                    var attribute = fragments[1].Split(new[] { '[', ']' }, 3);

                    current = this.FilterByClass(this.SelectCandidates(current, children, sibling), fragments[0], attribute[0], attribute.ElementAtOrDefault(1));
                    children = false;
                    sibling = false;
                }
                else if (part.Contains('#'))
                {
                    var fragments = part.Split(new[] { '#' }, 2);
                    var attribute = fragments[1].Split(new[] { '[', ']' }, 3);

                    current = this.FilterById(this.SelectCandidates(current, children, sibling), fragments[0], attribute[0], attribute.ElementAtOrDefault(1));
                    children = false;
                    sibling = false;
                }
                else if (part == ">")
                {
                    children = true;
                    sibling = false;
                }
                else if (part == "+")
                {
                    children = false;
                    sibling = true;
                }
                else
                {
                    var attribute = part.Split(new[] { '[', ']' }, 3);

                    current = this.FilterByTag(this.SelectCandidates(current, children, sibling), attribute[0], attribute.ElementAtOrDefault(1));
                    children = false;
                    sibling = false;
                }
            }

            return current.Distinct();
        }

        private IEnumerable<HTag> SelectCandidates(IEnumerable<HTag> source, bool children, bool sibling)
        {
            if (children)
                return source.SelectMany(x => x.Children.OfType<HTag>()).Distinct();

            if (sibling)
                return source.Select(x => x.Nexts<HTag>().FirstOrDefault()).Where(x => x != null);

            return source.SelectMany(x => x.Descendants<HTag>()).Distinct();
        }

        private static bool MatchAttribute(HTag element, string attribute)
        {
            if (String.IsNullOrEmpty(attribute))
                return true;

            if (attribute.Contains("~="))
            {
                var fragments = attribute.Split(new[] { '~', '=' }, StringSplitOptions.RemoveEmptyEntries);
                var name = fragments[0];
                var value = fragments[1].Trim('\'');

                return element.Attributes[name].Split(' ').Contains(value);
            }

            if (attribute.Contains("|="))
            {
                var fragments = attribute.Split(new[] { '|', '=' }, StringSplitOptions.RemoveEmptyEntries);
                var name = fragments[0];
                var value = fragments[1].Trim('\'', '"');

                return element.Attributes[name].Split('-').Contains(value);
            }

            if (attribute.Contains("="))
            {
                var fragments = attribute.Split('=');
                var name = fragments[0];
                var value = fragments[1].Trim('\'');

                return element.Attributes[name] == value;
            }

            return element.Attributes.Any(x => x.Name == attribute);
        }

        private IEnumerable<HTag> FilterByClass(IEnumerable<HTag> source, string tagname, string classname, string attribute)
        {
            return from element in source
                   where !this.skips.Intersect(element.Ancestors()).Any()
                   where String.IsNullOrEmpty(tagname) || element.Tag == tagname
                   let classes = element.Attributes[HAttributeType.Class].Split(' ')
                   where classes.Contains(classname)
                   where HQuery.MatchAttribute(element, attribute)
                   select element;
        }

        private IEnumerable<HTag> FilterById(IEnumerable<HTag> source, string tagname, string id, string attribute)
        {
            return from element in source
                   where !this.skips.Intersect(element.Ancestors()).Any()
                   where String.IsNullOrEmpty(tagname) || element.Tag == tagname
                   where element.Attributes[HAttributeType.Id] == id
                   where HQuery.MatchAttribute(element, attribute)
                   select element;
        }

        private IEnumerable<HTag> FilterByTag(IEnumerable<HTag> source, string tagname, string attribute)
        {
            return from element in source
                   where !this.skips.Intersect(element.Ancestors()).Any()
                   where element.Tag == tagname
                   where HQuery.MatchAttribute(element, attribute)
                   select element;
        }

        public static HQuery On(HContainer container)
        {
            return new HQuery(container);
        }

        public HQuery Skip(params HContainer[] containers)
        {
            this.skips.AddRange(containers.Where(x => x != null));
            return this;
        }

        public IEnumerable<HTag> Execute(string selector)
        {
            this.AddParts(selector.Split(new[] { ' ' }, StringSplitOptions.RemoveEmptyEntries));
            return this.Execute();
        }
    }
}
