﻿using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Linq;
using System.Text;
using EFramework.Testing.Composition;
using NUnit.Framework;

namespace EFramework.Testing.Tests.Composition
{
    [TestFixture]
    public class ObjectFactoryFixture
    {
        #region Independent Object

        [Test]
        public void CanCreateIndependentObject()
        {
            var factory = new ObjectFactory();
            var instance = factory.Create<IndependentClass>();

            Assert.NotNull(instance, "The factored object should be not empty.");
        }

        private class IndependentClass
        {
        }

        #endregion

        #region Field Injections

        [Test]
        public void CanInjectField()
        {
            var factory = new ObjectFactory();
            factory.Register(new FieldInjectedReference());

            var instance = factory.Create<FieldInjectedClass>();

            Assert.NotNull(instance, "The factored object should be not empty.");
            instance.Validate();
        }

        private class FieldInjectedClass
        {
            [Import]
            public FieldInjectedReference PublicField;

            [Import]
            private FieldInjectedReference PrivateField;


            public void Validate()
            {
                Assert.NotNull(this.PublicField, "The public field should be injected.");
                Assert.NotNull(this.PrivateField, "The private field should be injected.");
            }
        }

        private class FieldInjectedReference
        {
        }

        #endregion

        #region Property Injections

        [Test]
        public void CanInjectProperty()
        {
            var factory = new ObjectFactory();
            factory.Register(new PropertyInjectedReference());

            var instance = factory.Create<PropertyInjectedClass>();

            Assert.NotNull(instance, "The factored object should be not empty.");
            instance.Validate();
        }

        private class PropertyInjectedClass
        {
            [Import]
            public PropertyInjectedReference PublicProperty { get; set; }

            [Import]
            private PropertyInjectedReference PrivateProperty { get; set; }


            public void Validate()
            {
                Assert.NotNull(this.PublicProperty, "The public property should be injected.");
                Assert.NotNull(this.PrivateProperty, "The private property should be injected.");
            }
        }

        private class PropertyInjectedReference
        {
        }

        #endregion

        #region MultiImport

        [Test]
        public void CanInjectArray()
        {
            var factory = new ObjectFactory();
            var reference1 = new ArrayInjectedClassReference();
            var reference2 = new ArrayInjectedClassReference();

            factory.Register(reference1, reference2);

            var instance = factory.Create<ArrayInjectedClass>();

            Assert.NotNull(instance, "The factored object should be not empty.");
            instance.Validate(new[] { reference1, reference2 });
        }

        private class ArrayInjectedClass
        {
            [ImportMany]
            private ArrayInjectedClassReference[] references;

            public void Validate(IEnumerable<ArrayInjectedClassReference> references)
            {
                Assert.NotNull(this.references, "The array should be not null.");
                CollectionAssert.AreEquivalent(references, this.references, "The injected collection should be equivalent of a given collection.");
            }
        }

        private class ArrayInjectedClassReference
        {
        }

        #endregion

        #region Lazy Injection

        [Test]
        public void CanInjectLazyField()
        {
            var factory = new ObjectFactory();
            factory.Register(new LazyInjectedClassReference());

            var instance = factory.Create<LazyInjectedClass>();

            Assert.NotNull(instance, "The factored object should be not empty.");
            instance.Validate();
        }

        private class LazyInjectedClass
        {
            [Import]
            private Lazy<LazyInjectedClassReference> field;


            public void Validate()
            {
                Assert.NotNull(this.field, "The field should be injected.");
                Assert.False(this.field.IsValueCreated, "The value should not be initially created.");
                Assert.NotNull(this.field.Value, "The lazy field should provide the value.");
            }
        }

        private class LazyInjectedClassReference
        { 
        }

        #endregion

        #region Satisfaction

        [Test]
        public void CanInvokeSatisfactionMethod()
        {
            var factory = new ObjectFactory();
            var instance = factory.Create<SatisfiedInjectedClass>();

            Assert.NotNull(instance, "The factored object should be not empty.");
            instance.Validate();
        }

        private class SatisfiedInjectedClass : IPartImportsSatisfiedNotification
        {
            private bool invoked;


            void IPartImportsSatisfiedNotification.OnImportsSatisfied()
            {
                this.invoked = true;
            }


            public void Validate()
            {
                Assert.True(this.invoked, "The IPartImportsSatisfiedNotification.OnImportsSatisfied should be invoked.");
            }
        }

        #endregion
    }
}
