﻿using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Text;

namespace EFramework.Testing.Composition
{
    public class ObjectFactory
    {
        private readonly Dictionary<Type, object[]> instances;


        public ObjectFactory()
        {
            this.instances = new Dictionary<Type, object[]>();
        }


        public T Create<T>()
        {
            var instance = this.CreateInstance<T>();

            this.InjectFields(typeof(T), instance);
            this.InjectProperties(typeof(T), instance);
            this.Satisfy(instance);

            return instance;
        }

        public void Register<T>(params T[] instances)
        {
            this.instances.Add(typeof(T), instances.OfType<object>().ToArray());
        }


        private T CreateInstance<T>()
        {
            return Activator.CreateInstance<T>();
        }

        private void InjectFields(Type type, object instance)
        {
            this.InjectImportFields(type, instance);
            this.InjectImportManyFields(type, instance);
        }

        private void InjectImportFields(Type type, object instance)
        {
            var fields = from field in type.GetFields(BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance)
                         where field.IsDefined(typeof(ImportAttribute), false)
                         select field;

            foreach (var field in fields)
            {
                if (field.FieldType.IsGenericType && field.FieldType.GetGenericTypeDefinition().Equals(typeof(Lazy<>)))
                {
                    var target = field.FieldType.GetGenericArguments().First();
                    var call = Expression.Constant(this.instances[target].First());
                    var lambda = Expression.Lambda(typeof(Func<>).MakeGenericType(target), call);
                    var constructor = field.FieldType.GetConstructor(new[] { typeof(Func<>).MakeGenericType(target) });
                    var lazy = constructor.Invoke(new[] { lambda.Compile() });

                    field.SetValue(instance, lazy);
                }
                else
                {
                    field.SetValue(instance, this.instances[field.FieldType].First());
                }
            }
        }

        private void InjectImportManyFields(Type type, object instance)
        {
            var fields = from field in type.GetFields(BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance)
                         where field.IsDefined(typeof(ImportManyAttribute), false)
                         select field;

            foreach (var field in fields)
            {
                if (field.FieldType.IsArray)
                {
                    var source = this.instances[field.FieldType.GetElementType()];
                    var array = Array.CreateInstance(field.FieldType.GetElementType(), source.Length);

                    Array.Copy(source, array, source.Length);
                    field.SetValue(instance, array);
                }
            }
        }

        private void InjectProperties(Type type, object instance)
        {
            var properties = from property in type.GetProperties(BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance)
                             where property.IsDefined(typeof(ImportAttribute), false)
                             select property;

            foreach (var property in properties)
            {
                property.SetValue(instance, this.instances[property.PropertyType].First(), null);
            }
        }

        private void Satisfy(object instance)
        {
            if (instance is IPartImportsSatisfiedNotification == true)
            {
                (instance as IPartImportsSatisfiedNotification).OnImportsSatisfied();
            }
        }
    }
}
