﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Data;
using System.Windows.Input;
using System.Windows.Interactivity;

namespace MDownloader.Core.Client.Behaviours.Actions
{
    public class CommandAction : TargetedTriggerAction<FrameworkElement>, ICommandSource
    {
        public static readonly DependencyProperty CommandProperty =
            DependencyProperty.Register(
                "Command",
                typeof(ICommand),
                typeof(CommandAction),
                new PropertyMetadata(default(ICommand), CommandAction.OnCommandChanged));

        public static readonly DependencyProperty CommandParameterProperty =
            DependencyProperty.Register(
                "CommandParameter",
                typeof(object),
                typeof(CommandAction),
                new PropertyMetadata(CommandAction.OnCommandParameterChanged));

        public static readonly DependencyProperty CommandTargetProperty =
            DependencyProperty.Register(
                "CommandTarget",
                typeof(IInputElement),
                typeof(CommandAction),
                new PropertyMetadata());

        public static readonly DependencyProperty SyncOwnerIsEnabledProperty =
            DependencyProperty.Register(
                "SyncOwnerIsEnabled",
                typeof(bool),
                typeof(CommandAction),
                new PropertyMetadata());


        public ICommand Command
        {
            get { return (ICommand)this.GetValue(CommandAction.CommandProperty); }
            set { this.SetValue(CommandAction.CommandProperty, value); }
        }

        public object CommandParameter
        {
            get { return (object)this.GetValue(CommandAction.CommandParameterProperty); }
            set { this.SetValue(CommandAction.CommandParameterProperty, value); }
        }

        public IInputElement CommandTarget
        {
            get { return (IInputElement)this.GetValue(CommandAction.CommandTargetProperty); }
            set { this.SetValue(CommandAction.CommandTargetProperty, value); }
        }

        public bool SyncOwnerIsEnabled
        {
            get { return (bool)this.GetValue(CommandAction.SyncOwnerIsEnabledProperty); }
            set { this.SetValue(CommandAction.SyncOwnerIsEnabledProperty, value); }
        }


        private EventHandler CanExecuteChanged;

        private void OnCanExecuteChanged(object sender, EventArgs e)
        {
            this.UpdateCanExecute();
        }

        private static void OnCommandChanged(DependencyObject target, DependencyPropertyChangedEventArgs args)
        {
            ((CommandAction)target).OnCommandChanged((ICommand)args.OldValue, (ICommand)args.NewValue);
        }

        private static void OnCommandParameterChanged(DependencyObject target, DependencyPropertyChangedEventArgs args)
        {
            ((CommandAction)target).UpdateCanExecute();
        }

        protected override void Invoke(object o)
        {
            if (this.Command != null)
            {
                var binding = BindingOperations.GetBindingExpression(this, CommandAction.CommandParameterProperty);
                if (binding != null)
                {
                    binding.UpdateTarget();
                }

                var routedCommand = this.Command as RoutedCommand;
                if (routedCommand != null)
                {
                    routedCommand.Execute(this.CommandParameter, this.CommandTarget);
                }
                else
                {
                    this.Command.Execute(this.CommandParameter);
                }
            }
        }

        private void OnCommandChanged(ICommand oldCommand, ICommand newCommand)
        {
            if (oldCommand != null)
            {
                this.UnhookCommandCanExecuteChangedEventHandler(oldCommand);
            }

            if (newCommand != null)
            {
                this.HookupCommandCanExecuteChangedEventHandler(newCommand);
            }
        }

        private void HookupCommandCanExecuteChangedEventHandler(ICommand command)
        {
            this.CanExecuteChanged = new EventHandler(this.OnCanExecuteChanged);
            command.CanExecuteChanged += this.CanExecuteChanged;
            this.UpdateCanExecute();
        }

        private void UnhookCommandCanExecuteChangedEventHandler(ICommand command)
        {
            command.CanExecuteChanged -= CanExecuteChanged;
            this.UpdateCanExecute();
        }

        private void UpdateCanExecute()
        {
            if (this.Command != null)
            {
                var routedCommand = this.Command as RoutedCommand;
                if (routedCommand != null)
                {
                    this.IsEnabled = routedCommand.CanExecute(this.CommandParameter, this.CommandTarget);
                }
                else
                {
                    this.IsEnabled = this.Command.CanExecute(this.CommandParameter);
                }

                if (this.Target != null && this.SyncOwnerIsEnabled)
                {
                    this.Target.IsEnabled = IsEnabled;
                }
            }
        }
    }
}
