﻿using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Linq;
using System.Text;
using EFramework.Core.Contracts;
using EFramework.Core.Events;
using EFramework.Model;
using MDownloader.Main.Entities;
using MDownloader.Main.Interfaces;
using MDownloader.Main.Server.Captcher.Interfaces;
using System.Threading;
using EFramework.Model.Tracking;

namespace MDownloader.Main.Server.Captcher.Implementation
{
    [Export(typeof(ICaptcherInternal))]
    [PartCreationPolicy(CreationPolicy.Shared)]
    internal class Captcher : ICaptcherInternal
    {
        private readonly Container container;
        private readonly Dictionary<Guid, ICaptcha> captchas;
        private readonly EventsManager eventsManager;


        public Captcher()
        {
            this.container = new Container(Guid.NewGuid());
            this.captchas = new Dictionary<Guid, ICaptcha>();
            this.eventsManager = new EventsManager(this);
        }


        private const string CaptchaAddedEvent = "CaptchaAdded";
        public event EventHandler<CaptchaEventArgs> CaptchaAdded
        {
            add { this.eventsManager.Add(Captcher.CaptchaAddedEvent, value); }
            remove { this.eventsManager.Remove(Captcher.CaptchaAddedEvent, value); }
        }

        private void OnCaptchaAdded(params Captcha[] captchas)
        {
            this.eventsManager.Invoke(Captcher.CaptchaAddedEvent, new CaptchaEventArgs(captchas));
        }

        private const string CaptchaRemovedEvent = "CaptchaRemoved";
        public event EventHandler<CaptchaEventArgs> CaptchaRemoved
        {
            add { this.eventsManager.Add(Captcher.CaptchaRemovedEvent, value); }
            remove { this.eventsManager.Remove(Captcher.CaptchaRemovedEvent, value); }
        }

        private void OnCaptchaRemoved(params Captcha[] captchas)
        {
            this.eventsManager.Invoke(Captcher.CaptchaRemovedEvent, new CaptchaEventArgs(captchas));
        }


        public void Process(Task task, ICaptcha captcha)
        {
            Ensure.IsNotNull(() => task);
            Ensure.IsNotNull(() => captcha);

            var id = Guid.NewGuid();
            var entity = new Captcha(id) 
            { 
                Data = captcha.GetImage() 
            };

            entity.Add(task.Identifier);
            this.captchas.Add(id, captcha);

            this.container.Add(entity);
            this.OnCaptchaAdded(entity);
            this.HookUp(task, entity, captcha);
        }

        public void Complete(Guid captchaId, string text)
        {
            var captcha = this.container.Get<Captcha>(captchaId);
            captcha.Text = text;
        }

        private void HookUp(Task task, Captcha entity, ICaptcha captcha)
        {
            var synchronizer = new ManualResetEvent(false);
            var handler = default(EventHandler<TrackerEventArgs>);
            var left = 180;
            
            handler = (sender, args) =>
            {
                if (args.Owner == entity && args.PropertyName == CaptchaProperties.Text)
                {
                    synchronizer.Set();
                }
            };

            this.container.Changed += handler;

            try
            {
                while (left > 0)
                {
                    if (synchronizer.WaitOne(TimeSpan.FromSeconds(1)) == true)
                    {
                        if (String.IsNullOrEmpty(entity.Text) == false)
                        {
                            captcha.Satisfy(entity.Text);
                        }

                        break;
                    }
                    else
                    {
                        left--;
                    }

                    task.Counter = left;
                }
            }
            finally
            {
                task.Counter = 0;

                this.container.Changed -= handler;

                this.container.Remove(entity);
                this.OnCaptchaRemoved(entity);

                synchronizer.Close();
                synchronizer.Dispose();
                synchronizer = null;
            }
        }
    }
}
