﻿using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Linq;
using System.Text;
using System.Transactions;
using EFramework.Model;
using EFramework.Model.Merging;
using EFramework.Model.Tracking;
using MDownloader.Main.Entities;
using MDownloader.Main.Server.Model.Interfaces;
using MDownloader.Main.Server.Repository;
using MDownloader.Main.Server.Repository.Extensions;
using MDownloader.Main.Server.Repository.Interfaces;
using MDownloader.Main.Server.Writer.Interfaces;

namespace MDownloader.Main.Server.Model.Implementation
{
    [Export(typeof(IModelInternal))]
    [PartCreationPolicy(CreationPolicy.Shared)]
    internal class Model : IModelInternal, IRemoveNotifier
    {
        /// <summary>
        /// Holds a reference to repository.
        /// </summary>
        [Import]
        private IRepositoryInternal repository;

        [Import]
        private IWriterInternal writer;

        /// <summary>
        /// Holds a reference to current model container.
        /// </summary>
        private Container container;


        public Model()
        {
            this.container = new Container(Guid.NewGuid());
            this.container.Changed += this.OnChanged;
        }


        private void OnChanged(object sender, TrackerEventArgs e)
        {
            if (e.Owner is Link)
            {
                this.OnLinkChanged(e.Owner as Link, e.PropertyName);
            }

            if (e.Owner is Task)
            {
                this.OnTaskChanged(e.Owner as Task, e.PropertyName);
            }
        }


        public event EventHandler<LinkEventArgs> LinkAdded;
        private void OnLinkAdded(params Link[] links)
        {
            var handler = this.LinkAdded;
            var args = new LinkEventArgs(links);

            if (handler != null)
                handler(this, args);
        }

        public event EventHandler<LinkEventArgs> LinkChanged;
        private void OnLinkChanged(Link link, string propertyName)
        {
            var handler = this.LinkChanged;
            var args = new LinkEventArgs(link, propertyName);

            if (handler != null)
                handler(this, args);
        }

        public event EventHandler<LinkEventArgs> LinkRemoved;
        private void OnLinkRemoved(params Link[] links)
        {
            var handler = this.LinkRemoved;
            var args = new LinkEventArgs(links);

            if (handler != null)
                handler(this, args);
        }


        public event EventHandler<TaskEventArgs> TaskAdded;
        private void OnTaskAdded(params Task[] links)
        {
            var handler = this.TaskAdded;
            var args = new TaskEventArgs(links);

            if (handler != null)
                handler(this, args);
        }

        public event EventHandler<TaskEventArgs> TaskChanged;
        private void OnTaskChanged(Task link, string propertyName)
        {
            var handler = this.TaskChanged;
            var args = new TaskEventArgs(link, propertyName);

            if (handler != null)
                handler(this, args);
        }

        public event EventHandler<TaskEventArgs> TaskRemoved;
        private void OnTaskRemoved(params Task[] links)
        {
            var handler = this.TaskRemoved;
            var args = new TaskEventArgs(links);

            if (handler != null)
                handler(this, args);
        }


        public void Add(Package<Link>[] packages)
        {
            using (var transaction = new TransactionScope(TransactionScopeOption.RequiresNew))
            {
                using (var session = this.repository.OpenSession())
                {
                    foreach (var package in packages)
                    {
                        session.Persist(package);
                    }
                }

                transaction.Complete();
            }

            var mergeSettings = new MergeSettings { CloneEntities = false };
            foreach (var package in packages)
            {
                this.container.Merge(package, mergeSettings);
            }

            var query = from package in packages
                        select package.Leader();

            this.OnLinkAdded(query.ToArray());
        }

        public void Add(Package<Task>[] packages)
        {
            var mergeSettings = new MergeSettings { CloneEntities = false };
            foreach (var package in packages)
            {
                this.container.Merge(package, mergeSettings);
            }

            var query = from package in packages
                        select package.Leader();

            this.OnTaskAdded(query.ToArray());
        }

        public void Remove(Guid[] resources)
        {
            var resourceRemover = new ResourceRemover(this.container);
            var collectedResources = resourceRemover.Collect(resources);

            using (var transaction = new TransactionScope(TransactionScopeOption.RequiresNew))
            {
                using (var session = this.repository.OpenSession())
                {
                    session.Delete(collectedResources);
                }

                transaction.Complete();
            }

            resourceRemover.Remove();
            resourceRemover.Notify(this);
        }

        public void Remove(Task task)
        {
            this.container.Remove(task);
            this.OnTaskRemoved(task);
        }


        public File[] GetFiles()
        {
            return this.container.GetAll<File>().ToArray();
        }

        public Link[] GetLinks(Guid[] resources)
        {
            return this.container.Get<Link>(resources).ToArray();
        }

        public LinkPath GetPath(Link link)
        {
            var result = new LinkPath { Link = link };

            var urlId = link.GetSingle<Url>();
            result.Url = this.container.Get<Url>(urlId);

            var fileId = link.GetSingle<File>();
            result.File = this.container.Get<File>(fileId);

            var releaseId = link.GetSingleOrDefault<Release>();
            if (releaseId != null)
            {
                result.Release = this.container.Get<Release>(releaseId);
            }

            return result;
        }


        public void Initialize()
        {
            var mergeSettings = new MergeSettings { CloneEntities = false };

            var files = this.repository.GetFiles();
            foreach (var file in files)
            {
                this.container.Merge(file, mergeSettings);
            }

            foreach (var file in this.container.GetAll<File>())
            {
                var multiStream = this.writer.Get(file.Location);
                var multiStreamInfo = multiStream.GetInfo();

                file.Progress = multiStreamInfo.Size();

                if (file.SizeExactness == true && file.Progress == file.Size)
                {
                }
            }
        }

        public void Flush(IFlushSink flushSink)
        {
            flushSink.OnLinkAdded(this.container.GetAll<Link>().ToArray());
        }


        void IRemoveNotifier.OnLinkRemoved(params Link[] links)
        {
            this.OnLinkRemoved(links);
        }
    }
}
