﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using MDownloader.Main.Entities;
using MDownloader.Main.Interfaces;
using MDownloader.Main.Interfaces.Base;
using MDownloader.Main.Server.Captcher.Interfaces;
using MDownloader.Main.Server.Executor;
using MDownloader.Main.Server.Scheduler.Interfaces;
using MDownloader.Main.Server.Writer.Interfaces;

namespace MDownloader.Main.Server.Scheduler
{
    internal class DownloadTask : TaskBase<ISchedulerContext>, IExecutorTask<ISchedulerContext>, IDisposable
    {
        private readonly IProvider provider;
        private readonly IMultiStream multiStream;
        private readonly ICaptcherInternal captcher;

        private readonly Url url;
        private readonly Link link;
        private readonly File file;
        private readonly Task task;


        public DownloadTask(IProvider provider, IMultiStream multiStream, ICaptcherInternal captcher, Task task, File file, Link link, Url url)
        {
            this.provider = provider;
            this.multiStream = multiStream;
            this.captcher = captcher;

            this.url = url;
            this.link = link;
            this.file = file;
            this.task = task;
        }


        public Link Link
        {
            get { return this.link; }
        }

        public Task Task
        {
            get { return this.task; }
        }

        protected override void Execute(ISchedulerContext context)
        {
            using (var session = this.provider.OpenSession(context.Browser))
            {
                var resource = session.Download(this.url.Target);

                this.SatisfyFeedbacks(resource);
                using (var data = resource.GetData())
                {
                    this.SatisfyFeedbacks(data);
                    using (var stream = data.OpenStream())
                    {
                        this.task.Status = Task.TaskStatus.Downloading;
                        this.multiStream.Write(stream, data.Offset, data.Count, tick =>
                        {
                            this.task.Progress = tick.Progress;
                            this.task.Speed = tick.Speed;
                        });
                    }
                }
            }
        }

        private void SatisfyFeedbacks(IProviderFeedbackProvider feedbackProvider)
        {
            feedbackProvider.Initialize();

            while (true)
            {
                var feedback = feedbackProvider.GetFeedback();
                if (feedback == null || feedback.IsReady())
                {
                    break;
                }

                if (feedback.IsWaitingRequired())
                {
                    feedback.SatisfyWaitingRequirement(this.OnSatisfyWaiting);
                }

                if (feedback.IsCaptchaRequired())
                {
                    feedback.SatisfyCaptchaRequirement(this.OnSatisfyCaptcha);
                }
            }
        }

        private void OnSatisfyWaiting(TimeSpan period)
        {
            var seconds = Convert.ToInt32(Math.Ceiling(period.TotalSeconds));
            var oneSecond = TimeSpan.FromSeconds(1);

            try
            {
                this.task.Status = Task.TaskStatus.Waiting;

                while (seconds > 0)
                {
                    this.task.Counter = seconds;
                    Thread.Sleep(oneSecond);

                    seconds--;
                }
            }
            finally
            {
                this.task.Counter = 0;
                this.task.Status = Task.TaskStatus.Working;
            }
        }

        private void OnSatisfyCaptcha(ICaptcha captcha)
        {
            try
            {
                this.task.Status = Task.TaskStatus.Captcha;
                this.captcher.Process(this.task, captcha);
            }
            finally
            {
                this.task.Status = Task.TaskStatus.Working;
            }
        }


        public void Dispose()
        {
        }
    }
}
