﻿using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Linq;
using System.Text;
using System.Threading;
using EFramework.Model;
using MDownloader.Main.Entities;
using MDownloader.Main.Interfaces;
using MDownloader.Main.Server.Browser.Interfaces;
using MDownloader.Main.Server.Executor.Interfaces;
using MDownloader.Main.Server.Model.Interfaces;
using MDownloader.Main.Server.Providers.Interfaces;
using MDownloader.Main.Server.Scheduler.Interfaces;
using MDownloader.Main.Server.Writer.Interfaces;
using MDownloader.Main.Server.Captcher.Interfaces;

namespace MDownloader.Main.Server.Scheduler.Implementation
{
    [Export(typeof(ISchedulerInternal))]
    [PartCreationPolicy(CreationPolicy.Shared)]
    internal class Scheduler : ISchedulerInternal, IWaiter
    {
        [Import]
        private IModelInternal model;

        [Import]
        private IExecutor executor;

        [Import]
        private IWriterInternal writer;

        [Import]
        private IBrowserFactoryInternal browserFactory;

        [Import]
        private IProviderServiceInternal providerService;

        [Import]
        private ICaptcherInternal captcher;

        private Queue queue;


        public Scheduler()
        {
            this.queue = new Queue();
            this.queue.Ready += this.OnSchedulerReady;
        }


        public void Queue(Link[] links)
        {
            foreach (var link in links)
            {
                var linkPath = this.model.GetPath(link);
                var metadata = this.providerService.Classify(linkPath.Url.Target);
                var provider = this.providerService.GetProvider(metadata);
                var scheduler = this.providerService.GetScheduler(metadata);

                var package = new Package<Task>();
                var task = new Task(Guid.NewGuid())
                {
                    Provider = metadata.Name,
                    Name = linkPath.Url.Target,
                    Size = linkPath.File.Size,
                    Progress = 0L,
                    Speed = 0L,
                    Status = Task.TaskStatus.Idle,
                };

                package.Add(task);
                task.Add(link.Identifier);

                var multiStream = this.writer.Get(linkPath.File.Location);
                var run = new DownloadTask(provider, multiStream, this.captcher, task, linkPath.File, linkPath.Link, linkPath.Url);

                link.Status = Link.LinkStatus.Queued;
                run.Started += (sender, args) => this.model.Add(new[] { package });
                run.Started += this.OnTaskStarted;
                run.Finished += this.OnTaskFinished;

                this.queue.Add(scheduler, run);
            }
        }


        private void OnTaskStarted(object sender, EventArgs e)
        {
            var task = sender as DownloadTask;

            task.Link.Status = Link.LinkStatus.Working;
            task.Task.Status = Task.TaskStatus.Working;
        }

        private void OnTaskFinished(object sender, EventArgs e)
        {
            var task = sender as DownloadTask;

            task.Link.Status = Link.LinkStatus.Completed;
            task.Task.Status = Task.TaskStatus.Idle;           

            this.model.Remove(task.Task);
            task.Dispose();
        }

        private void OnSchedulerReady(object sender, ReadyEventArgs e)
        {
            var task = e.Tasks.First();
            var context = new Context
            { 
                Browser = this.browserFactory.CreateBrowser(), 
                Waiter = this
            };

            task.Finished += (_, args) =>
            {
                e.Token.Release();
            };

            this.queue.Remove(e.Scheduler, task);
            this.executor.Queue<ISchedulerContext>(context, task);
        }


        void IWaiter.Wait(TimeSpan period)
        {
            Thread.Sleep(period);
        }
    }
}
