﻿using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Linq;
using System.ServiceModel;
using System.Text;
using EFramework.Model;
using EFramework.Model.Merging;
using EFramework.Model.Serialization;
using EFramework.Model.Splitting;
using EFramework.Model.Transfer;
using MDownloader.Main.Entities;
using MDownloader.Main.Interfaces;
using MDownloader.Main.Server.Matcher.Interfaces;
using MDownloader.Main.Server.Model.Interfaces;
using MDownloader.Main.Server.Parser.Interfaces;
using MDownloader.Main.Server.Services.Implementation;

namespace MDownloader.Main.Server.Services.Endpoints
{
    [Export(typeof(IDownloaderScheduleService))]
    [PartCreationPolicy(CreationPolicy.Shared)]
    internal class DownloaderScheduleService : IDownloaderScheduleService
    {
        /// <summary>
        /// Holds a reference to uri prober.
        /// </summary>
        [Import]
        private IProber prober;

        /// <summary>
        /// Holds a reference to text parser.
        /// </summary>
        [Import]
        private IParserInternal parser;

        /// <summary>
        /// Holds a reference to matcher.
        /// </summary>
        [Import]
        private IMatcherInternal matcher;

        /// <summary>
        /// Holds a reference to model.
        /// </summary>
        [Import]
        private IModelInternal model;


        /// <summary>
        /// Probes all links found in the given text.
        /// </summary>
        /// <param name="text">The text to search in.</param>
        /// <returns>Probe result.</returns>
        public ContainerType Probe(string text)
        {
            var serializer = new Serializer();
            var serializerSettings = new SerializerSettings { PreserveTrackingStates = true };

            var uris = this.parser.FindLinks(text);
            var urls = uris.Select(x => new ProberUrl { Target = x }).ToArray();
            var links = this.prober.Probe(urls);
            var packages = links.Select(this.GetPackage).ToArray();

            var outputContainer = new Container(Guid.NewGuid());
            var mergeSettings = new MergeSettings { CloneEntities = false };
            outputContainer.Merge(packages, mergeSettings);

            var containerType = serializer.Serialize(outputContainer, serializerSettings);
            return containerType;
        }

        /// <summary>
        /// Probes all the given urls.
        /// </summary>
        /// <param name="containerType">The all urls containers.</param>
        /// <returns>Probe result.</returns>
        public ContainerType Probe(ContainerType containerType)
        {
            var serializer = new Serializer();
            var serializerSettings = new SerializerSettings { PreserveTrackingStates = true };

            var inputContainer = serializer.Deserialize(containerType, serializerSettings);
            var view = new ProbeListView(inputContainer);
            var urls = view.GetUrls().Select(x => new ProberUrl { Target = x.Target }).ToArray();
            var links = this.prober.Probe(urls);
            var packages = links.Select(this.GetPackage).ToArray();

            var outputContainer = new Container(Guid.NewGuid());
            var mergeSettings = new MergeSettings { CloneEntities = false };
            outputContainer.Merge(packages, mergeSettings);

            return serializer.Serialize(outputContainer, serializerSettings);
        }

        /// <summary>
        /// Queues all links, file and packages.
        /// </summary>
        /// <param name="containerType">The artifacts to queue.</param>
        public void Queue(ContainerType containerType)
        {
            var serializer = new Serializer();
            var serializerSettings = new SerializerSettings { PreserveTrackingStates = true };
            var container = serializer.Deserialize(containerType, serializerSettings);

            var splitSettings = new SplitSettings { CloneEntities = true };
            var packages = container.Split<Link>(splitSettings).ToArray();

            this.matcher.AssignFiles(packages);
            this.model.Add(packages);
        }


        private Package<Link> GetPackage(IProberLink link)
        {
            var package = new Package<Link>();
            var urlEntity = new Url(Guid.NewGuid())
            {
                Target = link.Target,
            };

            var linkEntity = new Link(Guid.NewGuid())
            {
                Name = link.Name,
                Provider = link.Provider,
                Description = link.Description,
                Size = link.Size,
                SizeExactness = link.SizeExactness,
            };

            package.Add(linkEntity);
            package.Add(urlEntity);

            linkEntity.Add(urlEntity.Identifier);
            urlEntity.Add(linkEntity.Identifier);

            return package;
        }
    }
}
