﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using MDownloader.Main.Server.Writer.Interfaces;

namespace MDownloader.Main.Server.Writer.Implementation
{
    internal class MultiStreamMap : IMultiStreamLockOwner, IMultiStreamInfo
    {
        private readonly string location;
        private readonly Dictionary<string, FileStream> parts;
        private readonly HashSet<IMultiStreamLock> locks;
        private readonly List<MultiStreamPoint> points;


        public MultiStreamMap(string location)
        {
            this.location = location;
            this.parts = new Dictionary<string, FileStream>();
            this.locks = new HashSet<IMultiStreamLock>();
            this.points = new List<MultiStreamPoint>();

            this.Initialize();
        }


        private void Initialize()
        {
            var folder = Path.GetDirectoryName(this.location);
            var name = Path.GetFileName(this.location);

            if (Directory.Exists(folder) == true)
            {
                var files = Directory.GetFiles(folder, name + "*", SearchOption.TopDirectoryOnly);

                foreach (var file in files)
                {
                    this.AddFile(file);
                }
            }
        }


        public IMultiStreamLock Lock(long offset, long count)
        {
            var multiStreamLock = new MultiStreamLock(this, offset, count);
            this.locks.Add(multiStreamLock);
            return multiStreamLock;
        }

        public MultiStreamRange[] List(IMultiStreamLock multiStreamLock)
        {
            var query = from point in this.points
                        where multiStreamLock.Start <= point.Offset
                        where point.Offset < multiStreamLock.Start + multiStreamLock.Count
                        orderby point.Offset
                        select point;

            var first = query.FirstOrDefault();
            if (first == null || first.Offset > multiStreamLock.Start)
            {
                this.AddFile(multiStreamLock.Start);
            }

            var points = query.ToArray();
            var ranges = new List<MultiStreamRange>();

            for (int i = 0; i < points.Length; i++)
            {
                var location = points[i].Location;
                var stream = this.GetStream(location);

                if (i == points.Length - 1)
                {
                    ranges.Add(new MultiStreamRange(location, stream, points[i].Offset, multiStreamLock.Start + multiStreamLock.Count));
                }
                else
                {
                    ranges.Add(new MultiStreamRange(location, stream, points[i].Offset, points[i + 1].Offset - points[i].Offset));
                }
            }

            return ranges.ToArray();
        }

        public void Flush(MultiStreamRange[] ranges)
        {
            foreach (var range in ranges)
            {
                var stream = this.parts[range.Location];
                if (stream != null)
                {
                    stream.Flush();
                    stream.Close();
                    stream.Dispose();

                    this.parts[range.Location] = null;
                }
            }
        }


        private FileStream GetStream(string filename)
        {
            var path = Path.GetDirectoryName(filename);
            if (Directory.Exists(path) == false)
            {
                Directory.CreateDirectory(path);
            }

            if (this.parts[filename] == null && File.Exists(filename))
            {
                this.parts[filename] = new FileStream(filename, FileMode.Open, FileAccess.ReadWrite, FileShare.None, 16 * 1024, FileOptions.None);
            }
            else if (this.parts[filename] == null && File.Exists(filename) == false)
            {
                this.parts[filename] = new FileStream(filename, FileMode.Create, FileAccess.ReadWrite, FileShare.None, 16 * 1024, FileOptions.None);
            }

            return this.parts[filename];
        }

        private void AddFile(string file)
        {
            var suffix = file.Substring(this.location.Length);
            var offset = String.IsNullOrWhiteSpace(suffix) ? 0 : Int32.Parse(suffix);
            var point = new MultiStreamPoint(file, offset);

            this.parts.Add(file, null);
            this.points.Add(point);
        }

        private void AddFile(long offset)
        {
            var file = this.location + (offset == 0 ? String.Empty : "." + offset);
            var point = new MultiStreamPoint(file, offset);

            this.parts.Add(file, null);
            this.points.Add(point);
        }


        void IMultiStreamLockOwner.Release(IMultiStreamLock multiStreamLock)
        {
            this.locks.Remove(multiStreamLock);
        }
        
        bool IMultiStreamInfo.Exists()
        {
            return this.parts.Count > 0;
        }

        long IMultiStreamInfo.Size()
        {
            var size = 0L;

            foreach (var part in this.parts)
            {
                if (part.Value != null)
                {
                    size += part.Value.Length;
                }
                else
                {
                    size += new FileInfo(part.Key).Length;
                }
            }

            return size;
        }
    }
}
