﻿using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Linq;
using System.Text;
using EFramework.Model;
using EFramework.Model.Cloning;
using MDownloader.Main.Interfaces;
using MDownloader.Search.Entities;
using MDownloader.Search.Interfaces;
using MDownloader.Search.Server.Engines.Interfaces;
using MDownloader.Search.Server.Model.Interfaces;
using MDownloader.Search.Server.Scheduler.Interfaces;

namespace MDownloader.Search.Server.Scheduler.Implementation
{
    [Export(typeof(ISchedulerInternal))]
    internal class Scheduler : ISchedulerInternal
    {
        [Import]
        private IModelInternal model;

        [Import]
        private IExecutor executor;

        [Import]
        private IProber prober;

        [Import]
        private IBrowserFactory browserFactory;

        [Import]
        private IEngineServiceInternal engineService;


        public void Queue(Specification specification)
        {
            var parameters = this.model.GetParameters(specification);
            var engine = engineService.GetEngine();

            var cloneSettings = new CloneSettings();
            var clonedSpecification = specification.Clone(cloneSettings);

            var task = new SearchTask(engine, clonedSpecification, parameters);

            task.Finished += this.OnTaskFinished;

            this.executor.Queue(this.browserFactory.CreateBrowser(), task);
        }

        private void Queue(Specification specification, IEnumerable<Result> results)
        {
            var query = from result in results
                        from url in this.model.GetUrls(result)
                        select new ProberUrl
                        {
                            Target = url.Target,
                        };

            var probes = this.prober.Probe(query.ToArray());

            var merge = from result in results                        
                        let links = from url in this.model.GetUrls(result)
                                    join link in probes on url.Target equals link.Target
                                    select link
                        select new
                        {
                            Result = result,
                            Links = links.ToArray(),
                        };

            foreach (var entry in merge)
            {
                if (entry.Links.Any())
                {
                    entry.Result.Size = entry.Links.Sum(x => x.Size);
                    entry.Result.SizeExactness = entry.Links.All(x => x.SizeExactness);
                }
            }
        }


        private void OnTaskFinished(object sender, EventArgs e)
        {
            var task = sender as SearchTask;

            var specification = task.GetSpecification();
            var package = new Package<Specification>(specification.Identifier);
            package.Add(specification);

            var bag = task.GetBag();
            var results = new List<Result>();

            foreach (var searchResult in bag.GetResults())
            {
                var result = new Result(Guid.NewGuid())
                {
                    Name = searchResult.Name,
                    Provider = searchResult.Provider,
                    Files = searchResult.Links.Length,
                    Status = Result.ResultStatus.Idle,
                };

                foreach (var link in searchResult.Links)
                {
                    var url = new Url(Guid.NewGuid())
                    {
                        Target = link,
                    };

                    package.Add(url);
                    result.Add(url.Identifier);
                    url.Add(result.Identifier);
                }

                package.Add(result);
                results.Add(result);

                specification.Add(result.Identifier);
                result.Add(specification.Identifier);
            }

            this.model.AddResults(package);
            this.Queue(specification, results);
        }
    }
}
