﻿using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.ComponentModel.Composition.Hosting;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Windows;

namespace MDownloader.Core
{
    [Export(typeof(IServiceProvider))]
    internal class Bootstrapper : IServiceProvider
    {
        private static readonly CompositionContainer container;

        static Bootstrapper()
        {
            var catalog = new AggregateCatalog();
            var folder = Uri.UnescapeDataString(new Uri(typeof(Bootstrapper).Assembly.CodeBase).AbsolutePath);

            catalog.Catalogs.Add(new AssemblyCatalog(typeof(Bootstrapper).Assembly));
            catalog.Catalogs.Add(new DirectoryCatalog(Path.GetDirectoryName(folder), "MDownloader.Client.dll")); 
            catalog.Catalogs.Add(new DirectoryCatalog(Path.GetDirectoryName(folder), "MDownloader.Client.*.dll"));
            catalog.Catalogs.Add(new DirectoryCatalog(Path.GetDirectoryName(folder), "MDownloader.*.Client.dll"));

            Bootstrapper.container = new CompositionContainer(catalog);
        }

        public static void Initialize()
        {
            Bootstrapper.InitializeTemplates();
            Bootstrapper.InitializeWindow();
        }

        private static void InitializeTemplates()
        {
            var viewModels = from part in Bootstrapper.container.Catalog.Parts
                             from export in part.ExportDefinitions
                             from metadata in export.Metadata
                             where metadata.Key == CompositionConstants.ExportTypeIdentityMetadataName
                             where metadata.Value.ToString().EndsWith("ViewModel", StringComparison.InvariantCulture)
                             select metadata.Value.ToString();

            var viewPairs = from assembly in AppDomain.CurrentDomain.GetAssemblies()
                            from type in assembly.GetTypes()
                            join viewModelType in viewModels on type.FullName equals viewModelType
                            let viewType = assembly.GetType(type.FullName.Substring(0, type.FullName.Length - 5), false, false)
                            where viewType != null
                            select new
                            {
                                ViewType = viewType,
                                ViewModelType = type,
                            };

            var resourceDictionary = new ResourceDictionary();
            resourceDictionary.BeginInit();

            foreach (var viewPair in viewPairs)
            {
                var frameworkElementFactory = new FrameworkElementFactory(viewPair.ViewType);
                var dataTemplate = new DataTemplate { DataType = viewPair.ViewModelType, VisualTree = frameworkElementFactory };
                var dataTemplateKey = new DataTemplateKey(viewPair.ViewModelType);

                resourceDictionary.Add(dataTemplateKey, dataTemplate);
            }

            resourceDictionary.EndInit();
            App.Current.Resources.MergedDictionaries.Add(resourceDictionary);
        }

        private static void InitializeWindow()
        {
            App.Current.MainWindow = Bootstrapper.container.GetExportedValue<ShellView>();

            App.Current.MainWindow.Show();
            App.Current.ShutdownMode = ShutdownMode.OnMainWindowClose;
        }

        #region IServiceProvider Members

        object IServiceProvider.GetService(Type serviceType)
        {
            var typeInfo = Bootstrapper.container.GetType();
            var methodInfo = typeInfo.GetMethod("GetExportedValue", new Type[] { }).MakeGenericMethod(serviceType);
            var foundInstance = methodInfo.Invoke(Bootstrapper.container, null);

            return foundInstance;
        }

        #endregion
    }
}
