﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Moq;
using NUnit.Framework;
using QImpl.Core.Exceptions;
using QImpl.Core.Exceptions.UnexpectedData;

namespace QImpl.Core.Tests.Exceptions
{
    [TestFixture]
    public class UnexpectedDataExceptionFactoryFixture
    {
        [Test]
        public void CanCreateEmptyException()
        {
            var exception = UnexpectedDataExceptionFactory.CreateException<Object>();

            Assert.NotNull(exception, "Exception cannot be null.");
            Assert.NotNull(exception.Data, "Attachments cannot be null.");

            Assert.AreEqual(typeof(Object), exception.Response, "The response type should be properly set.");
            Assert.AreEqual(0, exception.Data.Count, "There should not be any attachment.");
        }

        [Test]
        public void CanProperlySetResponseType()
        {
            var exception = UnexpectedDataExceptionFactory.CreateException<CustomResponse>();

            Assert.NotNull(exception, "Exception cannot be null.");
            Assert.AreEqual(typeof(CustomResponse), exception.Response, "The response type should be properly set.");
        }

        private class CustomResponse
        {
        }

        [Test]
        public void CanProperlyDetectName()
        {
            var exception = UnexpectedDataExceptionFactory.CreateException<Object>(abc => abc.Create(String.Empty));

            Assert.NotNull(exception, "Exception cannot be null.");
            Assert.AreEqual(typeof(Object), exception.Response, "The response type should be properly set.");
            Assert.AreEqual(1, exception.Data.Count, "There should be exactly one attachment.");
            Assert.AreEqual("abc", exception.Data.Keys.First(), "The attachment name should be properly set.");
        }

        [Test]
        public void CanProperlyTransferBytes()
        {
            var data = new byte[] { 0x00, 0x01, 0x02 };
            var exception = UnexpectedDataExceptionFactory.CreateException<Object>(x => x.Create(data));

            Assert.NotNull(exception, "Exception cannot be null.");
            Assert.AreEqual(1, exception.Data.Count, "There should be exactly one attachment.");
            CollectionAssert.AreEqual(data, exception.Data.Values.First(), "The attachment data should be properly set.");
        }

        [Test]
        public void CanDealWithCustomArtifact()
        {
            var data = new byte[] { 0x00, 0x01, 0x02 };
            var artifact = new Mock<Artifact>();
            artifact.Setup(x => x.GetBytes()).Returns(data);

            var exception = UnexpectedDataExceptionFactory.CreateException<Object>(x => artifact.Object);

            Assert.NotNull(exception, "Exception cannot be null.");
            Assert.AreEqual(1, exception.Data.Count, "There should be exactly one attachment.");
            CollectionAssert.AreEqual(data, exception.Data.Values.First(), "The attachment data should be properly set.");
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanGuardArtifactParameter()
        {
            UnexpectedDataExceptionFactory.CreateException<Object>(null);
        }

        [Test, ExpectedException(typeof(ArgumentException))]
        public void CanDetectRepeatedParameterNames()
        {
            UnexpectedDataExceptionFactory.CreateException<Object>(x => x.Create(String.Empty), x => x.Create(String.Empty));
        }
    }
}
