﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using EFramework.Networking.Parser;
using System.Text.RegularExpressions;
using System.Globalization;

namespace QImpl.Filestube.Implementation.Responses
{
    internal class SearchResponseParser
    {
        // parts: 2   1 GB  date: 2011-12-24
        //    349 MB  date: 2011-12-25 
        private readonly static Regex SummaryRegex = new Regex(@"(\s*parts:\s*(?<parts>[0-9]+))?(\s*(?<value>[0-9]+)\s*(?<unit>(GB|MB|kB)))?(\s*date:\s*(?<date>[0-9]{4}-[0-9]{2}-[0-9]{2}))?");

        public SearchResponse Parse(HDocument page)
        {
            var response = new SearchResponse
            {
                Results = new List<SearchResponse.Entry>(),
            };

            this.ProcessResultsEntry(response.Results, page);
            this.ProcessNewResult(response.Results, page);

            return response;
        }


        private void ProcessResultsEntry(IList<SearchResponse.Entry> entries, HDocument document)
        {
            foreach (var entry in HQuery.On(document).Execute("div.fsResultsEntry"))
            {
                var element = new SearchResponse.Entry();

                this.ProcessName(element, HQuery.On(entry).Execute("a.resultsLink").FirstOrDefault());
                this.ProcessProvider(element, HQuery.On(entry).Execute("div.fsResultsDetails span").FirstOrDefault());
                this.ProcessExtension(element, HQuery.On(entry).Execute("div.fsResultsDetails b").FirstOrDefault());
                this.ProcessSummary(element, HQuery.On(entry).Execute("div.fsResultsDetails b").Take(1).SelectMany(x => x.Nexts<HText>()).FirstOrDefault());
                this.ProcessSource(element, HQuery.On(entry).Execute("div.fsUrl a[rel=nofollow]").FirstOrDefault());
                this.ProcessDetails(element, HQuery.On(entry).Execute("a.resultsLink").FirstOrDefault());

                entries.Add(element);
            }
        }

        private void ProcessNewResult(IList<SearchResponse.Entry> entries, HDocument document)
        {
            foreach (var entry in HQuery.On(document).Execute("div#newresult"))
            {
                var element = new SearchResponse.Entry();

                this.ProcessName(element, HQuery.On(entry).Execute("a.resultsLink").FirstOrDefault());
                this.ProcessProvider(element, HQuery.On(entry).Execute("div span b").ElementAtOrDefault(0));
                this.ProcessExtension(element, HQuery.On(entry).Execute("div span b").ElementAtOrDefault(1));
                this.ProcessSummary(element, HQuery.On(entry).Execute("div span b").Skip(1).Take(1).SelectMany(x => x.Nexts<HText>()).FirstOrDefault());
                this.ProcessSource(element, HQuery.On(entry).Execute("div a[rel=nofollow]").FirstOrDefault());
                this.ProcessDetails(element, HQuery.On(entry).Execute("a.resultsLink").FirstOrDefault());

                entries.Add(element);
            }
        }


        private void ProcessName(SearchResponse.Entry entry, HTag element)
        {
            if (element != null)
            {
                entry.Name = this.Join(element);
            }
        }

        private void ProcessProvider(SearchResponse.Entry entry, HTag element)
        {
            if (element != null)
            {
                entry.Provider = this.Join(element);
            }
        }

        private void ProcessExtension(SearchResponse.Entry entry, HTag element)
        {
            if (element != null)
            {
                entry.Extension = this.Join(element);
            }
        }

        private void ProcessSummary(SearchResponse.Entry entry, HText element)
        {
            if (element != null)
            {
                var summary = element.Value;
                var match = SearchResponseParser.SummaryRegex.Match(summary);

                var parts = match.Groups["parts"].Value;
                if (String.IsNullOrWhiteSpace(parts) == false)
                {
                    entry.Parts = Int32.Parse(parts.Trim());
                }

                var value = match.Groups["value"].Value;
                var unit = match.Groups["unit"].Value;
                if (String.IsNullOrWhiteSpace(value) == false && String.IsNullOrWhiteSpace(unit) == false)
                {
                    var multiplier = 1024L;

                    if (unit == "MB") multiplier = 1024 * 1024;
                    if (unit == "GB") multiplier = 1024 * 1024 * 1024;

                    entry.SizeExactness = false;
                    entry.Size = multiplier * Int32.Parse(value);
                }

                var date = match.Groups["date"].Value;
                if (String.IsNullOrWhiteSpace(date) == false)
                {
                    entry.Date = DateTime.ParseExact(date.Trim(), "yyyy-MM-dd", CultureInfo.InvariantCulture);
                }
            }
        }

        private void ProcessSource(SearchResponse.Entry entry, HTag element)
        {
            if (element != null)
            {
                entry.Source = this.Join(element);
            }
        }

        private void ProcessDetails(SearchResponse.Entry entry, HTag element)
        {
            if (element != null)
            {
                entry.Details = element.Attributes[HAttributeType.Hyperlink];
            }
        }

        
        private string Join(HTag tag)
        {
            var query = from element in tag.Descendants<HText>()
                        select element.Value;

            return String.Join(String.Empty, query);
        }
    }
}
