﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using EFramework.Networking.Parser;
using QImpl.Responses.Document;

namespace QImpl.Hotfile.Implementation.Responses.Processors
{
    /// <summary>
    /// A processor to find files outputed by file checker.
    /// </summary>
    internal class FilesProcessor : DocumentProcessorList<CheckFilesResponse, IEnumerable<HTag>, HTag>
    {
        private readonly UrlParser parser = new UrlParser();
        private readonly Regex sizeRegex = new Regex("^(?<size>[0-9]+(\\.[0-9]+)?)\\s*(?<multiplier>(kB|Mb|Gb))$", RegexOptions.IgnoreCase);


        protected override bool Validate(IEnumerable<HTag> element)
        {
            return element.Any() == true;
        }

        protected override Func<HDocument, IEnumerable<HTag>> GetQuery()
        {
            return document => HQuery.On(document).Execute("table.tbl");
        }

        protected override void Initialize(CheckFilesResponse response)
        {
            response.Files = new List<CheckFilesResponse.Entry>();
        }

        protected override IEnumerable<HTag> GetRows(IEnumerable<HTag> element)
        {
            foreach (var table in element)
            {
                var headerQuery = from header in HQuery.On(table).Execute("tr td b")
                                  from text in header.Children.OfType<HText>()
                                  where String.Equals(text.Value, "results", StringComparison.InvariantCultureIgnoreCase) == true
                                  select header;

                if (headerQuery.Any() == true)
                {
                    foreach (var row in HQuery.On(table).Execute("tr"))
                    {
                        yield return row;
                    }
                }
            }
        }

        protected override void Execute(CheckFilesResponse response, HTag row)
        {
            var element = new CheckFilesResponse.Entry();

            this.ProcessAnchor(element, HQuery.On(row).Execute("td a[href]").FirstOrDefault());
            this.ProcessAnchorRow(element, HQuery.On(row).Execute("td").ElementAtOrDefault(0));
            this.ProcessSize(element, HQuery.On(row).Execute("td").ElementAtOrDefault(1));
            this.ProcessStatus(element, HQuery.On(row).Execute("td[nowrap] span").FirstOrDefault());

            if (element.Validate() == true)
            {
                response.Files.Add(element); ;
            }
        }


        private void ProcessAnchor(CheckFilesResponse.Entry entry, HTag element)
        {
            if (element != null)
            {
                var uriText = element.Attributes[HAttributeType.Hyperlink];
                var uri = default(Uri);

                if (Uri.TryCreate(uriText, UriKind.Absolute, out uri))
                {
                    entry.Id = this.parser.GetIdentifier(uri);
                    entry.Name = this.parser.GetName(uri);
                }
            }
        }

        private void ProcessAnchorRow(CheckFilesResponse.Entry entry, HTag element)
        {
            if (element != null)
            {
                var text = element.Descendants<HText>().FirstOrDefault();
                if (text != null)
                {
                    var uriText = text.Value.Trim();
                    var uri = default(Uri);

                    if (Uri.TryCreate(uriText, UriKind.Absolute, out uri))
                    {
                        entry.Id = this.parser.GetIdentifier(uri);
                        entry.Name = this.parser.GetName(uri);
                    }
                }
            }
        }

        private void ProcessSize(CheckFilesResponse.Entry entry, HTag element)
        {
            if (element != null)
            {
                var text = element.Children.OfType<HText>().Select(x => x.Value).FirstOrDefault();
                if (String.IsNullOrEmpty(text) == false)
                {
                    var match = this.sizeRegex.Match(text.TrimEnd());
                    if (match.Success == true)
                    {
                        var sizeText = match.Groups["size"].Value;
                        var sizeValue = Decimal.Parse(sizeText, NumberStyles.AllowDecimalPoint, new NumberFormatInfo { NumberDecimalSeparator = "." });
                        var multiplierText = match.Groups["multiplier"].Value;
                        var multiplierValue = 1;

                        switch (multiplierText.ToLower())
                        {
                            case "kb": multiplierValue = 1024; break;
                            case "mb": multiplierValue = 1024 * 1024; break;
                            case "gb": multiplierValue = 1024 * 1024 * 1024; break;
                        }

                        entry.Size = (long)Math.Round(multiplierValue * sizeValue);
                    }
                }
            }
        }

        private void ProcessStatus(CheckFilesResponse.Entry entry, HTag element)
        {
            if (element != null)
            {
                var text = element.Children.OfType<HText>().Select(x => x.Value).FirstOrDefault();

                if (String.IsNullOrEmpty(text) == false && text.Equals("Existent", StringComparison.InvariantCultureIgnoreCase))
                {
                    entry.Status = CheckFilesResponse.Status.FileOk;
                }
                else
                {
                    entry.Status = CheckFilesResponse.Status.NotFound;
                }
            }
        }

    }
}
