﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using EFramework.Core.Contracts;
using EFramework.Networking.Parser;
using MDownloader.Main.Interfaces.Exceptions;
using QImpl.Core.Exceptions;
using QImpl.Megaupload.Implementation.Exceptions;

namespace QImpl.Megaupload.Implementation.Responses
{
    internal class FileInfoResponseParser
    {
        private readonly Regex sizeRegex = new Regex("^(?<size>[0-9]+(\\.[0-9]+)?)\\s*(?<multiplier>(kB|MB|GB))$", RegexOptions.IgnoreCase);


        public FileInfoResponse Parse(HDocument document)
        {
            Ensure.IsNotNull(() => document);

            var result = new FileInfoResponse();

            this.EnsureAvailability(HQuery.On(document).Execute("div.main_bg").FirstOrDefault());
            this.ProcessName(result, HQuery.On(document).Execute("div.download_file_name").FirstOrDefault());
            this.ProcessSize(result, HQuery.On(document).Execute("div.download_file_size").FirstOrDefault());
            this.ProcessLink(result, HQuery.On(document).Execute("a#dlbutton").FirstOrDefault());

            if (result.Validate() == false)
            {
                throw UnexpectedDataExceptionFactory.CreateException<FileInfoResponseParser>(page => page.Create(document));
            }

            return result;
        }


        private void ProcessName(FileInfoResponse response, HTag element)
        {
            if (element != null)
            {
                response.Name = element.Children.OfType<HText>().Select(x => x.Value.Trim()).FirstOrDefault();
            }
        }

        private void ProcessSize(FileInfoResponse response, HTag element)
        {
            if (element != null)
            {
                var text = element.Children.OfType<HText>().Select(x => x.Value).FirstOrDefault();
                if (String.IsNullOrEmpty(text) == false)
                {
                    var match = this.sizeRegex.Match(text.Trim());
                    if (match.Success == true)
                    {
                        var sizeText = match.Groups["size"].Value;
                        var sizeValue = Decimal.Parse(sizeText, NumberStyles.AllowDecimalPoint, new NumberFormatInfo { NumberDecimalSeparator = "." });
                        var multiplierText = match.Groups["multiplier"].Value;
                        var multiplierValue = 1;

                        switch (multiplierText.ToLower())
                        {
                            case "kb": multiplierValue = 1024; break;
                            case "mb": multiplierValue = 1024 * 1024; break;
                            case "gb": multiplierValue = 1024 * 1024 * 1024; break;
                        }

                        response.Size = (long)Math.Round(multiplierValue * sizeValue);
                    }
                }
            }
        }

        private void ProcessLink(FileInfoResponse response, HTag element)
        {
            if (element != null)
            {
                var href = element.Attributes[HAttributeType.Hyperlink];
                if (String.IsNullOrWhiteSpace(href) == false)
                {
                    response.Link = href;
                }
            }
        }

        private void EnsureAvailability(HTag element)
        {
            if (element != null)
            {
                if (HQuery.On(element).Execute("div.na_text_block").Any() == true)
                {
                    throw new MegauploadException(MegauploadReason.LinkNotAvailable, "The link is not available.");
                }

                if (HQuery.On(element).Execute("div.access_na_text").Any() == true)
                {
                    throw new MegauploadException(MegauploadReason.LinkNotAvailable, "The link is temporarily not available.");
                }
            }
        }
    }
}
