﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using EFramework.Core.Contracts;
using EFramework.Networking;
using EFramework.Networking.Parser;
using MDownloader.Main.Interfaces.Exceptions;
using QImpl.Core.Exceptions;
using QImpl.Megaupload.Implementation.Exceptions;

namespace QImpl.Megaupload.Implementation.Responses
{
    internal class StreamResponseParser
    {
        public StreamResponse Parse(Stream stream, NetworkCallback callback)
        {
            Ensure.IsNotNull(() => stream);
            Ensure.IsNotNull(() => callback);

            if (this.IsBinary(callback) == true)
            {
                var result = new StreamResponse();

                this.ProcessAttachment(result, callback);
                this.ProcessSize(result, callback);
                result.Data = stream;

                if (result.Validate() == false)
                {
                    throw UnexpectedDataExceptionFactory.CreateException<StreamResponseParser>(networkCallback => networkCallback.Create(callback));
                }

                return result;
            }

            var data = default(string);

            if (this.IsHtml(callback) == true)
            {
                this.EnsureLoaded(stream, ref data);

                var document = HDocumentFactory.Create(data);
                var meta = HQuery.On(document).Execute("meta[http-equiv='Refresh']");
                var message = HQuery.On(document).Execute("body").SelectMany(x => x.Descendants<HText>()).FirstOrDefault();

                if (meta.Any() && message != null && String.IsNullOrWhiteSpace(message.Value) == false)
                {
                    throw new MegauploadException(MegauploadReason.SessionNotAvailable, message.Value);
                }
            }

            try
            {
                this.EnsureLoaded(stream, ref data);

                throw UnexpectedDataExceptionFactory.CreateException<StreamResponseParser>(networkCallback => networkCallback.Create(callback), streamData => streamData.Create(data));
            }
            finally
            {
                stream.Close();
                stream.Dispose();
            }
        }


        private void ProcessAttachment(StreamResponse response, NetworkCallback callback)
        {
            Ensure.IsNotNull(() => response);
            Ensure.IsNotNull(() => callback);

            if (NetworkParser.HasContentDisposition(callback) == true)
            {
                response.Attachment = NetworkParser.GetContentDispositionAttachment(callback);
            }
        }

        private void ProcessSize(StreamResponse response, NetworkCallback callback)
        {
            Ensure.IsNotNull(() => response);
            Ensure.IsNotNull(() => callback);

            if (NetworkParser.HasContentLength(callback) == true)
            {
                response.Size = NetworkParser.GetContentLength(callback);
            }
        }

        private bool IsBinary(NetworkCallback callback)
        {
            Ensure.IsNotNull(() => callback);

            if (NetworkParser.HasContentType(callback) == true)
            {
                var type = NetworkParser.GetContentType(callback);
                switch (type)
                {
                    case @"application/octet-stream": return true;
                }
            }

            return false;
        }

        private bool IsHtml(NetworkCallback callback)
        {
            Ensure.IsNotNull(() => callback);

            if (NetworkParser.HasContentType(callback) == true)
            {
                var type = NetworkParser.GetContentType(callback);
                switch (type)
                {
                    case @"text/html": return true;
                }
            }

            return false;
        }

        private void EnsureLoaded(Stream stream, ref string data)
        {
            Ensure.IsNotNull(() => stream);

            if (data == null)
            {
                using (var inner = stream)
                using (var reader = new StreamReader(inner))
                {
                    data = reader.ReadToEnd();
                }
            }
        }
    }
}
